#!/usr/bin/env python

"""
This module tests multiple message cyclic send tasks.
"""
import time
import unittest

import can

from .config import TEST_INTERFACE_SOCKETCAN


@unittest.skipUnless(TEST_INTERFACE_SOCKETCAN, "skip testing of socketcan")
class CyclicSocketCan(unittest.TestCase):
    BITRATE = 500000
    TIMEOUT = 0.1

    INTERFACE_1 = "socketcan"
    CHANNEL_1 = "vcan0"
    INTERFACE_2 = "socketcan"
    CHANNEL_2 = "vcan0"

    PERIOD = 1.0

    DELTA = 0.01

    def _find_start_index(self, tx_messages, message):
        """
        :param tx_messages:
            The list of messages that were passed to the periodic backend
        :param message:
            The message whose data we wish to match and align to

        :returns: start index in the tx_messages
        """
        start_index = -1
        for index, tx_message in enumerate(tx_messages):
            if tx_message.data == message.data:
                start_index = index
                break
        return start_index

    def setUp(self):
        self._send_bus = can.Bus(
            interface=self.INTERFACE_1, channel=self.CHANNEL_1, bitrate=self.BITRATE
        )
        self._recv_bus = can.Bus(
            interface=self.INTERFACE_2, channel=self.CHANNEL_2, bitrate=self.BITRATE
        )

    def tearDown(self):
        self._send_bus.shutdown()
        self._recv_bus.shutdown()

    def test_cyclic_initializer_list(self):
        messages = []
        messages.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x11, 0x11, 0x11, 0x11, 0x11, 0x11],
                is_extended_id=False,
            )
        )
        messages.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x22, 0x22, 0x22, 0x22, 0x22, 0x22],
                is_extended_id=False,
            )
        )
        messages.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x33, 0x33, 0x33, 0x33, 0x33, 0x33],
                is_extended_id=False,
            )
        )
        messages.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x44, 0x44, 0x44, 0x44, 0x44, 0x44],
                is_extended_id=False,
            )
        )
        messages.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x55, 0x55, 0x55, 0x55, 0x55, 0x55],
                is_extended_id=False,
            )
        )

        task = self._send_bus.send_periodic(messages, self.PERIOD)
        self.assertIsInstance(task, can.broadcastmanager.CyclicSendTaskABC)

        results = []
        for _ in range(len(messages) * 2):
            result = self._recv_bus.recv(self.PERIOD * 2)
            if result:
                results.append(result)

        task.stop()

        # Find starting index for each
        start_index = self._find_start_index(messages, results[0])
        self.assertTrue(start_index != -1)

        # Now go through the partitioned results and assert that they're equal
        for rx_index, rx_message in enumerate(results):
            tx_message = messages[start_index]

            self.assertIsNotNone(rx_message)
            self.assertEqual(tx_message.arbitration_id, rx_message.arbitration_id)
            self.assertEqual(tx_message.dlc, rx_message.dlc)
            self.assertEqual(tx_message.data, rx_message.data)
            self.assertEqual(tx_message.is_extended_id, rx_message.is_extended_id)
            self.assertEqual(tx_message.is_remote_frame, rx_message.is_remote_frame)
            self.assertEqual(tx_message.is_error_frame, rx_message.is_error_frame)
            self.assertEqual(tx_message.is_fd, rx_message.is_fd)

            start_index = (start_index + 1) % len(messages)

    def test_cyclic_initializer_tuple(self):
        messages = []
        messages.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x11, 0x11, 0x11, 0x11, 0x11, 0x11],
                is_extended_id=False,
            )
        )
        messages.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x22, 0x22, 0x22, 0x22, 0x22, 0x22],
                is_extended_id=False,
            )
        )
        messages.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x33, 0x33, 0x33, 0x33, 0x33, 0x33],
                is_extended_id=False,
            )
        )
        messages.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x44, 0x44, 0x44, 0x44, 0x44, 0x44],
                is_extended_id=False,
            )
        )
        messages.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x55, 0x55, 0x55, 0x55, 0x55, 0x55],
                is_extended_id=False,
            )
        )
        messages = tuple(messages)

        self.assertIsInstance(messages, tuple)

        task = self._send_bus.send_periodic(messages, self.PERIOD)
        self.assertIsInstance(task, can.broadcastmanager.CyclicSendTaskABC)

        results = []
        for _ in range(len(messages) * 2):
            result = self._recv_bus.recv(self.PERIOD * 2)
            if result:
                results.append(result)

        task.stop()

        # Find starting index for each
        start_index = self._find_start_index(messages, results[0])
        self.assertTrue(start_index != -1)

        # Now go through the partitioned results and assert that they're equal
        for rx_index, rx_message in enumerate(results):
            tx_message = messages[start_index]

            self.assertIsNotNone(rx_message)
            self.assertEqual(tx_message.arbitration_id, rx_message.arbitration_id)
            self.assertEqual(tx_message.dlc, rx_message.dlc)
            self.assertEqual(tx_message.data, rx_message.data)
            self.assertEqual(tx_message.is_extended_id, rx_message.is_extended_id)
            self.assertEqual(tx_message.is_remote_frame, rx_message.is_remote_frame)
            self.assertEqual(tx_message.is_error_frame, rx_message.is_error_frame)
            self.assertEqual(tx_message.is_fd, rx_message.is_fd)

            start_index = (start_index + 1) % len(messages)

    def test_cyclic_initializer_message(self):
        message = can.Message(
            arbitration_id=0x401,
            data=[0x11, 0x11, 0x11, 0x11, 0x11, 0x11],
            is_extended_id=False,
        )

        task = self._send_bus.send_periodic(message, self.PERIOD)
        self.assertIsInstance(task, can.broadcastmanager.CyclicSendTaskABC)

        # Take advantage of kernel's queueing mechanisms
        time.sleep(4 * self.PERIOD)
        task.stop()

        for _ in range(4):
            tx_message = message
            rx_message = self._recv_bus.recv(self.TIMEOUT)

            self.assertIsNotNone(rx_message)
            self.assertEqual(tx_message.arbitration_id, rx_message.arbitration_id)
            self.assertEqual(tx_message.dlc, rx_message.dlc)
            self.assertEqual(tx_message.data, rx_message.data)
            self.assertEqual(tx_message.is_extended_id, rx_message.is_extended_id)
            self.assertEqual(tx_message.is_remote_frame, rx_message.is_remote_frame)
            self.assertEqual(tx_message.is_error_frame, rx_message.is_error_frame)
            self.assertEqual(tx_message.is_fd, rx_message.is_fd)

    def test_cyclic_initializer_invalid_none(self):
        with self.assertRaises(ValueError):
            task = self._send_bus.send_periodic(None, self.PERIOD)

    def test_cyclic_initializer_invalid_empty_list(self):
        with self.assertRaises(ValueError):
            task = self._send_bus.send_periodic([], self.PERIOD)

    def test_cyclic_initializer_different_arbitration_ids(self):
        messages = []
        messages.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x11, 0x11, 0x11, 0x11, 0x11, 0x11],
                is_extended_id=False,
            )
        )
        messages.append(
            can.Message(
                arbitration_id=0x3E1,
                data=[0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE],
                is_extended_id=False,
            )
        )
        with self.assertRaises(ValueError):
            task = self._send_bus.send_periodic(messages, self.PERIOD)

    def test_start_already_started_task(self):
        messages_a = can.Message(
            arbitration_id=0x401,
            data=[0x11, 0x11, 0x11, 0x11, 0x11, 0x11],
            is_extended_id=False,
        )

        task_a = self._send_bus.send_periodic(messages_a, self.PERIOD)
        time.sleep(0.1)

        # Task restarting is permitted as of #1440
        task_a.start()
        task_a.stop()

    def test_create_same_id(self):
        messages_a = can.Message(
            arbitration_id=0x401,
            data=[0x11, 0x11, 0x11, 0x11, 0x11, 0x11],
            is_extended_id=False,
        )

        messages_b = can.Message(
            arbitration_id=0x401,
            data=[0x22, 0x22, 0x22, 0x22, 0x22, 0x22],
            is_extended_id=False,
        )

        task_a = self._send_bus.send_periodic(messages_a, self.PERIOD)
        self.assertIsInstance(task_a, can.broadcastmanager.CyclicSendTaskABC)
        task_b = self._send_bus.send_periodic(messages_b, self.PERIOD)
        self.assertIsInstance(task_b, can.broadcastmanager.CyclicSendTaskABC)

        time.sleep(self.PERIOD * 4)

        task_a.stop()
        task_b.stop()

        msgs = []
        for _ in range(4):
            msg = self._recv_bus.recv(self.PERIOD * 2)
            self.assertIsNotNone(msg)

            msgs.append(msg)

        self.assertTrue(len(msgs) >= 4)

        # Both messages should be recevied on the bus,
        # even with the same arbitration id
        msg_a_data_present = msg_b_data_present = False
        for rx_message in msgs:
            self.assertTrue(
                rx_message.arbitration_id
                == messages_a.arbitration_id
                == messages_b.arbitration_id
            )
            if rx_message.data == messages_a.data:
                msg_a_data_present = True
            if rx_message.data == messages_b.data:
                msg_b_data_present = True

        self.assertTrue(msg_a_data_present)
        self.assertTrue(msg_b_data_present)

    def test_modify_data_list(self):
        messages_odd = []
        messages_odd.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x11, 0x11, 0x11, 0x11, 0x11, 0x11],
                is_extended_id=False,
            )
        )
        messages_odd.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x33, 0x33, 0x33, 0x33, 0x33, 0x33],
                is_extended_id=False,
            )
        )
        messages_odd.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x55, 0x55, 0x55, 0x55, 0x55, 0x55],
                is_extended_id=False,
            )
        )
        messages_even = []
        messages_even.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x22, 0x22, 0x22, 0x22, 0x22, 0x22],
                is_extended_id=False,
            )
        )
        messages_even.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x44, 0x44, 0x44, 0x44, 0x44, 0x44],
                is_extended_id=False,
            )
        )
        messages_even.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x66, 0x66, 0x66, 0x66, 0x66, 0x66],
                is_extended_id=False,
            )
        )

        task = self._send_bus.send_periodic(messages_odd, self.PERIOD)
        self.assertIsInstance(task, can.broadcastmanager.ModifiableCyclicTaskABC)

        results_odd = []
        results_even = []
        for _ in range(len(messages_odd) * 2):
            result = self._recv_bus.recv(self.PERIOD * 2)
            if result:
                results_odd.append(result)

        task.modify_data(messages_even)
        for _ in range(len(messages_even) * 2):
            result = self._recv_bus.recv(self.PERIOD * 2)
            if result:
                results_even.append(result)

        task.stop()

        # Make sure we received some messages
        self.assertTrue(len(results_even) != 0)
        self.assertTrue(len(results_odd) != 0)

        # Find starting index for each
        start_index_even = self._find_start_index(messages_even, results_even[0])
        self.assertTrue(start_index_even != -1)

        start_index_odd = self._find_start_index(messages_odd, results_odd[0])
        self.assertTrue(start_index_odd != -1)

        # Now go through the partitioned results and assert that they're equal
        for rx_index, rx_message in enumerate(results_even):
            tx_message = messages_even[start_index_even]

            self.assertEqual(tx_message.arbitration_id, rx_message.arbitration_id)
            self.assertEqual(tx_message.dlc, rx_message.dlc)
            self.assertEqual(tx_message.data, rx_message.data)
            self.assertEqual(tx_message.is_extended_id, rx_message.is_extended_id)
            self.assertEqual(tx_message.is_remote_frame, rx_message.is_remote_frame)
            self.assertEqual(tx_message.is_error_frame, rx_message.is_error_frame)
            self.assertEqual(tx_message.is_fd, rx_message.is_fd)

            start_index_even = (start_index_even + 1) % len(messages_even)

            if rx_index != 0:
                prev_rx_message = results_even[rx_index - 1]
                # Assert timestamps are within the expected period
                self.assertTrue(
                    abs(
                        (rx_message.timestamp - prev_rx_message.timestamp) - self.PERIOD
                    )
                    <= self.DELTA
                )

        for rx_index, rx_message in enumerate(results_odd):
            tx_message = messages_odd[start_index_odd]

            self.assertEqual(tx_message.arbitration_id, rx_message.arbitration_id)
            self.assertEqual(tx_message.dlc, rx_message.dlc)
            self.assertEqual(tx_message.data, rx_message.data)
            self.assertEqual(tx_message.is_extended_id, rx_message.is_extended_id)
            self.assertEqual(tx_message.is_remote_frame, rx_message.is_remote_frame)
            self.assertEqual(tx_message.is_error_frame, rx_message.is_error_frame)
            self.assertEqual(tx_message.is_fd, rx_message.is_fd)

            start_index_odd = (start_index_odd + 1) % len(messages_odd)

            if rx_index != 0:
                prev_rx_message = results_odd[rx_index - 1]
                # Assert timestamps are within the expected period
                self.assertTrue(
                    abs(
                        (rx_message.timestamp - prev_rx_message.timestamp) - self.PERIOD
                    )
                    <= self.DELTA
                )

    def test_modify_data_message(self):
        message_odd = can.Message(
            arbitration_id=0x401,
            data=[0x11, 0x11, 0x11, 0x11, 0x11, 0x11],
            is_extended_id=False,
        )
        message_even = can.Message(
            arbitration_id=0x401,
            data=[0x22, 0x22, 0x22, 0x22, 0x22, 0x22],
            is_extended_id=False,
        )
        task = self._send_bus.send_periodic(message_odd, self.PERIOD)
        self.assertIsInstance(task, can.broadcastmanager.ModifiableCyclicTaskABC)

        results_odd = []
        results_even = []
        for _ in range(1 * 4):
            result = self._recv_bus.recv(self.PERIOD * 2)
            if result:
                results_odd.append(result)

        task.modify_data(message_even)
        for _ in range(1 * 4):
            result = self._recv_bus.recv(self.PERIOD * 2)
            if result:
                results_even.append(result)

        task.stop()

        # Now go through the partitioned results and assert that they're equal
        for rx_index, rx_message in enumerate(results_even):
            tx_message = message_even

            self.assertEqual(tx_message.arbitration_id, rx_message.arbitration_id)
            self.assertEqual(tx_message.dlc, rx_message.dlc)
            self.assertEqual(tx_message.data, rx_message.data)
            self.assertEqual(tx_message.is_extended_id, rx_message.is_extended_id)
            self.assertEqual(tx_message.is_remote_frame, rx_message.is_remote_frame)
            self.assertEqual(tx_message.is_error_frame, rx_message.is_error_frame)
            self.assertEqual(tx_message.is_fd, rx_message.is_fd)

            if rx_index != 0:
                prev_rx_message = results_even[rx_index - 1]
                # Assert timestamps are within the expected period
                self.assertTrue(
                    abs(
                        (rx_message.timestamp - prev_rx_message.timestamp) - self.PERIOD
                    )
                    <= self.DELTA
                )

        for rx_index, rx_message in enumerate(results_odd):
            tx_message = message_odd

            self.assertEqual(tx_message.arbitration_id, rx_message.arbitration_id)
            self.assertEqual(tx_message.dlc, rx_message.dlc)
            self.assertEqual(tx_message.data, rx_message.data)
            self.assertEqual(tx_message.is_extended_id, rx_message.is_extended_id)
            self.assertEqual(tx_message.is_remote_frame, rx_message.is_remote_frame)
            self.assertEqual(tx_message.is_error_frame, rx_message.is_error_frame)
            self.assertEqual(tx_message.is_fd, rx_message.is_fd)

            if rx_index != 0:
                prev_rx_message = results_odd[rx_index - 1]
                # Assert timestamps are within the expected period
                self.assertTrue(
                    abs(
                        (rx_message.timestamp - prev_rx_message.timestamp) - self.PERIOD
                    )
                    <= self.DELTA
                )

    def test_modify_data_invalid(self):
        message = can.Message(
            arbitration_id=0x401,
            data=[0x11, 0x11, 0x11, 0x11, 0x11, 0x11],
            is_extended_id=False,
        )
        task = self._send_bus.send_periodic(message, self.PERIOD)
        self.assertIsInstance(task, can.broadcastmanager.ModifiableCyclicTaskABC)

        time.sleep(2 * self.PERIOD)

        with self.assertRaises(ValueError):
            task.modify_data(None)

    def test_modify_data_unequal_lengths(self):
        message = can.Message(
            arbitration_id=0x401,
            data=[0x11, 0x11, 0x11, 0x11, 0x11, 0x11],
            is_extended_id=False,
        )
        new_messages = []
        new_messages.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x11, 0x11, 0x11, 0x11, 0x11, 0x11],
                is_extended_id=False,
            )
        )
        new_messages.append(
            can.Message(
                arbitration_id=0x401,
                data=[0x22, 0x22, 0x22, 0x22, 0x22, 0x22],
                is_extended_id=False,
            )
        )

        task = self._send_bus.send_periodic(message, self.PERIOD)
        self.assertIsInstance(task, can.broadcastmanager.ModifiableCyclicTaskABC)

        time.sleep(2 * self.PERIOD)

        with self.assertRaises(ValueError):
            task.modify_data(new_messages)

    def test_modify_data_different_arbitration_id_than_original(self):
        old_message = can.Message(
            arbitration_id=0x401,
            data=[0x11, 0x11, 0x11, 0x11, 0x11, 0x11],
            is_extended_id=False,
        )
        new_message = can.Message(
            arbitration_id=0x3E1,
            data=[0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE],
            is_extended_id=False,
        )

        task = self._send_bus.send_periodic(old_message, self.PERIOD)
        self.assertIsInstance(task, can.broadcastmanager.ModifiableCyclicTaskABC)

        time.sleep(2 * self.PERIOD)

        with self.assertRaises(ValueError):
            task.modify_data(new_message)

    def test_stop_all_periodic_tasks_and_remove_task(self):
        message_a = can.Message(
            arbitration_id=0x401,
            data=[0x11, 0x11, 0x11, 0x11, 0x11, 0x11],
            is_extended_id=False,
        )
        message_b = can.Message(
            arbitration_id=0x402,
            data=[0x22, 0x22, 0x22, 0x22, 0x22, 0x22],
            is_extended_id=False,
        )
        message_c = can.Message(
            arbitration_id=0x403,
            data=[0x33, 0x33, 0x33, 0x33, 0x33, 0x33],
            is_extended_id=False,
        )

        # Start Tasks
        task_a = self._send_bus.send_periodic(message_a, self.PERIOD)
        task_b = self._send_bus.send_periodic(message_b, self.PERIOD)
        task_c = self._send_bus.send_periodic(message_c, self.PERIOD)

        self.assertIsInstance(task_a, can.broadcastmanager.ModifiableCyclicTaskABC)
        self.assertIsInstance(task_b, can.broadcastmanager.ModifiableCyclicTaskABC)
        self.assertIsInstance(task_c, can.broadcastmanager.ModifiableCyclicTaskABC)

        for _ in range(6):
            _ = self._recv_bus.recv(self.PERIOD)

        # Stop all tasks and delete
        self._send_bus.stop_all_periodic_tasks(remove_tasks=True)

        # Now wait for a few periods, after which we should definitely not
        # receive any CAN messages
        time.sleep(4 * self.PERIOD)

        # If we successfully deleted everything, then we will eventually read
        # 0 messages.
        successfully_stopped = False
        for _ in range(6):
            rx_message = self._recv_bus.recv(self.PERIOD)

            if rx_message is None:
                successfully_stopped = True
                break
        self.assertTrue(successfully_stopped, "Still received messages after stopping")

        # None of the tasks should still be associated with the bus
        self.assertEqual(0, len(self._send_bus._periodic_tasks))


if __name__ == "__main__":
    unittest.main()



"""
Reader and writer for can logging files in peak trc format

See https://www.peak-system.com/produktcd/Pdf/English/PEAK_CAN_TRC_File_Format.pdf
for file format description

Version 1.1 will be implemented as it is most commonly used
"""  # noqa

import io
import logging
import os
from datetime import datetime, timedelta, timezone
from enum import Enum
from typing import Callable, Dict, Generator, List, Optional, TextIO, Union

from ..message import Message
from ..typechecking import StringPathLike
from ..util import channel2int, dlc2len, len2dlc
from .generic import FileIOMessageWriter, MessageReader

logger = logging.getLogger("can.io.trc")


class TRCFileVersion(Enum):
    UNKNOWN = 0
    V1_0 = 100
    V1_1 = 101
    V1_2 = 102
    V1_3 = 103
    V2_0 = 200
    V2_1 = 201

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented


class TRCReader(MessageReader):
    """
    Iterator of CAN messages from a TRC logging file.
    """

    file: TextIO

    def __init__(
        self,
        file: Union[StringPathLike, TextIO],
    ) -> None:
        """
        :param file: a path-like object or as file-like object to read from
                     If this is a file-like object, is has to opened in text
                     read mode, not binary read mode.
        """
        super().__init__(file, mode="r")
        self.file_version = TRCFileVersion.UNKNOWN
        self.start_time: Optional[datetime] = None
        self.columns: Dict[str, int] = {}

        if not self.file:
            raise ValueError("The given file cannot be None")

        self._parse_cols: Callable[[List[str]], Optional[Message]] = lambda x: None

    def _extract_header(self):
        line = ""
        for line in self.file:
            line = line.strip()
            if line.startswith(";$FILEVERSION"):
                logger.debug("TRCReader: Found file version '%s'", line)
                try:
                    file_version = line.split("=")[1]
                    if file_version == "1.1":
                        self.file_version = TRCFileVersion.V1_1
                    elif file_version == "2.0":
                        self.file_version = TRCFileVersion.V2_0
                    elif file_version == "2.1":
                        self.file_version = TRCFileVersion.V2_1
                    else:
                        self.file_version = TRCFileVersion.UNKNOWN
                except IndexError:
                    logger.debug("TRCReader: Failed to parse version")
            elif line.startswith(";$STARTTIME"):
                logger.debug("TRCReader: Found start time '%s'", line)
                try:
                    self.start_time = datetime(
                        1899, 12, 30, tzinfo=timezone.utc
                    ) + timedelta(days=float(line.split("=")[1]))
                except IndexError:
                    logger.debug("TRCReader: Failed to parse start time")
            elif line.startswith(";$COLUMNS"):
                logger.debug("TRCReader: Found columns '%s'", line)
                try:
                    columns = line.split("=")[1].split(",")
                    self.columns = {column: columns.index(column) for column in columns}
                except IndexError:
                    logger.debug("TRCReader: Failed to parse columns")
            elif line.startswith(";"):
                continue
            else:
                break

        if self.file_version >= TRCFileVersion.V1_1:
            if self.start_time is None:
                raise ValueError("File has no start time information")

        if self.file_version >= TRCFileVersion.V2_0:
            if not self.columns:
                raise ValueError("File has no column information")

        if self.file_version == TRCFileVersion.UNKNOWN:
            logger.info(
                "TRCReader: No file version was found, so version 1.0 is assumed"
            )
            self._parse_cols = self._parse_msg_V1_0
        elif self.file_version == TRCFileVersion.V1_0:
            self._parse_cols = self._parse_msg_V1_0
        elif self.file_version == TRCFileVersion.V1_1:
            self._parse_cols = self._parse_cols_V1_1
        elif self.file_version in [TRCFileVersion.V2_0, TRCFileVersion.V2_1]:
            self._parse_cols = self._parse_cols_V2_x
        else:
            raise NotImplementedError("File version not fully implemented for reading")

        return line

    def _parse_msg_V1_0(self, cols: List[str]) -> Optional[Message]:
        arbit_id = cols[2]
        if arbit_id == "FFFFFFFF":
            logger.info("TRCReader: Dropping bus info line")
            return None

        msg = Message()
        msg.timestamp = float(cols[1]) / 1000
        msg.arbitration_id = int(arbit_id, 16)
        msg.is_extended_id = len(arbit_id) > 4
        msg.channel = 1
        msg.dlc = int(cols[3])
        msg.data = bytearray([int(cols[i + 4], 16) for i in range(msg.dlc)])
        return msg

    def _parse_msg_V1_1(self, cols: List[str]) -> Optional[Message]:
        arbit_id = cols[3]

        msg = Message()
        if isinstance(self.start_time, datetime):
            msg.timestamp = (
                self.start_time + timedelta(milliseconds=float(cols[1]))
            ).timestamp()
        else:
            msg.timestamp = float(cols[1]) / 1000
        msg.arbitration_id = int(arbit_id, 16)
        msg.is_extended_id = len(arbit_id) > 4
        msg.channel = 1
        msg.dlc = int(cols[4])
        msg.data = bytearray([int(cols[i + 5], 16) for i in range(msg.dlc)])
        msg.is_rx = cols[2] == "Rx"
        return msg

    def _parse_msg_V2_x(self, cols: List[str]) -> Optional[Message]:
        type_ = cols[self.columns["T"]]
        bus = self.columns.get("B", None)

        if "l" in self.columns:
            length = int(cols[self.columns["l"]])
            dlc = len2dlc(length)
        elif "L" in self.columns:
            dlc = int(cols[self.columns["L"]])
            length = dlc2len(dlc)
        else:
            raise ValueError("No length/dlc columns present.")

        msg = Message()
        if isinstance(self.start_time, datetime):
            msg.timestamp = (
                self.start_time + timedelta(milliseconds=float(cols[self.columns["O"]]))
            ).timestamp()
        else:
            msg.timestamp = float(cols[1]) / 1000
        msg.arbitration_id = int(cols[self.columns["I"]], 16)
        msg.is_extended_id = len(cols[self.columns["I"]]) > 4
        msg.channel = int(cols[bus]) if bus is not None else 1
        msg.dlc = dlc
        msg.data = bytearray(
            [int(cols[i + self.columns["D"]], 16) for i in range(length)]
        )
        msg.is_rx = cols[self.columns["d"]] == "Rx"
        msg.is_fd = type_ in ["FD", "FB", "FE", "BI"]
        msg.bitrate_switch = type_ in ["FB", " FE"]
        msg.error_state_indicator = type_ in ["FE", "BI"]

        return msg

    def _parse_cols_V1_1(self, cols: List[str]) -> Optional[Message]:
        dtype = cols[2]
        if dtype in ("Tx", "Rx"):
            return self._parse_msg_V1_1(cols)
        else:
            logger.info("TRCReader: Unsupported type '%s'", dtype)
            return None

    def _parse_cols_V2_x(self, cols: List[str]) -> Optional[Message]:
        dtype = cols[self.columns["T"]]
        if dtype in ["DT", "FD", "FB"]:
            return self._parse_msg_V2_x(cols)
        else:
            logger.info("TRCReader: Unsupported type '%s'", dtype)
            return None

    def _parse_line(self, line: str) -> Optional[Message]:
        logger.debug("TRCReader: Parse '%s'", line)
        try:
            cols = line.split()
            return self._parse_cols(cols)
        except IndexError:
            logger.warning("TRCReader: Failed to parse message '%s'", line)
            return None

    def __iter__(self) -> Generator[Message, None, None]:
        first_line = self._extract_header()

        if first_line is not None:
            msg = self._parse_line(first_line)
            if msg is not None:
                yield msg

        for line in self.file:
            temp = line.strip()
            if temp.startswith(";"):
                # Comment line
                continue

            if len(temp) == 0:
                # Empty line
                continue

            msg = self._parse_line(temp)
            if msg is not None:
                yield msg

        self.stop()


class TRCWriter(FileIOMessageWriter):
    """Logs CAN data to text file (.trc).

    The measurement starts with the timestamp of the first registered message.
    If a message has a timestamp smaller than the previous one or None,
    it gets assigned the timestamp that was written for the last message.
    If the first message does not have a timestamp, it is set to zero.
    """

    file: TextIO
    first_timestamp: Optional[float]

    FORMAT_MESSAGE = (
        "{msgnr:>7} {time:13.3f} DT {channel:>2} {id:>8} {dir:>2} -  {dlc:<4} {data}"
    )
    FORMAT_MESSAGE_V1_0 = "{msgnr:>6}) {time:7.0f} {id:>8} {dlc:<1} {data}"

    def __init__(
        self,
        file: Union[StringPathLike, TextIO],
        channel: int = 1,
    ) -> None:
        """
        :param file: a path-like object or as file-like object to write to
                     If this is a file-like object, is has to opened in text
                     write mode, not binary write mode.
        :param channel: a default channel to use when the message does not
                        have a channel set
        """
        super().__init__(file, mode="w")
        self.channel = channel

        if isinstance(self.file, io.TextIOWrapper):
            self.file.reconfigure(newline="\r\n")
        else:
            raise TypeError("File must be opened in text mode.")

        self.filepath = os.path.abspath(self.file.name)
        self.header_written = False
        self.msgnr = 0
        self.first_timestamp = None
        self.file_version = TRCFileVersion.V2_1
        self._msg_fmt_string = self.FORMAT_MESSAGE_V1_0
        self._format_message = self._format_message_init

    def _write_header_V1_0(self, start_time: datetime) -> None:
        lines = [
            ";##########################################################################",
            f";   {self.filepath}",
            ";",
            ";    Generated by python-can TRCWriter",
            f";    Start time: {start_time}",
            ";    PCAN-Net: N/A",
            ";",
            ";    Columns description:",
            ";    ~~~~~~~~~~~~~~~~~~~~~",
            ";    +-current number in actual sample",
            ";    |     +time offset of message (ms",
            ";    |     |        +ID of message (hex",
            ";    |     |        |    +data length code",
            ";    |     |        |    |  +data bytes (hex ...",
            ";    |     |        |    |  |",
            ";----+- ---+--- ----+--- + -+ -- -- ...",
        ]
        self.file.writelines(line + "\n" for line in lines)

    def _write_header_V2_1(self, start_time: datetime) -> None:
        header_time = start_time - datetime(year=1899, month=12, day=30)
        lines = [
            ";$FILEVERSION=2.1",
            f";$STARTTIME={header_time/timedelta(days=1)}",
            ";$COLUMNS=N,O,T,B,I,d,R,L,D",
            ";",
            f";   {self.filepath}",
            ";",
            f";   Start time: {start_time}",
            ";   Generated by python-can TRCWriter",
            ";-------------------------------------------------------------------------------",
            ";   Bus   Name            Connection               Protocol",
            ";   N/A   N/A             N/A                      N/A",
            ";-------------------------------------------------------------------------------",
            ";   Message   Time    Type    ID     Rx/Tx",
            ";   Number    Offset  |  Bus  [hex]  |  Reserved",
            ";   |         [ms]    |  |    |      |  |  Data Length Code",
            ";   |         |       |  |    |      |  |  |    Data [hex] ...",
            ";   |         |       |  |    |      |  |  |    |",
            ";---+-- ------+------ +- +- --+----- +- +- +--- +- -- -- -- -- -- -- --",
        ]
        self.file.writelines(line + "\n" for line in lines)

    def _format_message_by_format(self, msg, channel):
        if msg.is_extended_id:
            arb_id = f"{msg.arbitration_id:07X}"
        else:
            arb_id = f"{msg.arbitration_id:04X}"

        data = [f"{byte:02X}" for byte in msg.data]

        serialized = self._msg_fmt_string.format(
            msgnr=self.msgnr,
            time=(msg.timestamp - self.first_timestamp) * 1000,
            channel=channel,
            id=arb_id,
            dir="Rx" if msg.is_rx else "Tx",
            dlc=msg.dlc,
            data=" ".join(data),
        )
        return serialized

    def _format_message_init(self, msg, channel):
        if self.file_version == TRCFileVersion.V1_0:
            self._format_message = self._format_message_by_format
            self._msg_fmt_string = self.FORMAT_MESSAGE_V1_0
        elif self.file_version == TRCFileVersion.V2_1:
            self._format_message = self._format_message_by_format
            self._msg_fmt_string = self.FORMAT_MESSAGE
        else:
            raise NotImplementedError("File format is not supported")

        return self._format_message_by_format(msg, channel)

    def write_header(self, timestamp: float) -> None:
        # write start of file header
        start_time = datetime.utcfromtimestamp(timestamp)

        if self.file_version == TRCFileVersion.V1_0:
            self._write_header_V1_0(start_time)
        elif self.file_version == TRCFileVersion.V2_1:
            self._write_header_V2_1(start_time)
        else:
            raise NotImplementedError("File format is not supported")
        self.header_written = True

    def log_event(self, message: str, timestamp: float) -> None:
        if not self.header_written:
            self.write_header(timestamp)

        self.file.write(message + "\n")

    def on_message_received(self, msg: Message) -> None:
        if self.first_timestamp is None:
            self.first_timestamp = msg.timestamp

        if msg.is_error_frame:
            logger.warning("TRCWriter: Logging error frames is not implemented")
            return

        if msg.is_remote_frame:
            logger.warning("TRCWriter: Logging remote frames is not implemented")
            return

        channel = channel2int(msg.channel)
        if channel is None:
            channel = self.channel
        else:
            # Many interfaces start channel numbering at 0 which is invalid
            channel += 1

        if msg.is_fd:
            logger.warning("TRCWriter: Logging CAN FD is not implemented")
            return
        else:
            serialized = self._format_message(msg, channel)
            self.msgnr += 1
        self.log_event(serialized, msg.timestamp)



"""
Interface for CANtact devices from Linklayer Labs
"""

import logging
import time
from typing import Any, Optional, Union
from unittest.mock import Mock

from can import BitTiming, BitTimingFd, BusABC, Message

from ..exceptions import (
    CanInitializationError,
    CanInterfaceNotImplementedError,
    error_check,
)
from ..util import check_or_adjust_timing_clock, deprecated_args_alias

logger = logging.getLogger(__name__)

try:
    import cantact
except ImportError:
    cantact = None
    logger.warning(
        "The CANtact module is not installed. Install it using `pip install cantact`"
    )


class CantactBus(BusABC):
    """CANtact interface"""

    @staticmethod
    def _detect_available_configs():
        try:
            interface = cantact.Interface()
        except (NameError, SystemError, AttributeError):
            logger.debug(
                "Could not import or instantiate cantact, so no configurations are available"
            )
            return []

        channels = []
        for i in range(0, interface.channel_count()):
            channels.append({"interface": "cantact", "channel": f"ch:{i}"})
        return channels

    @deprecated_args_alias(
        deprecation_start="4.2.0", deprecation_end="5.0.0", bit_timing="timing"
    )
    def __init__(
        self,
        channel: int,
        bitrate: int = 500_000,
        poll_interval: float = 0.01,
        monitor: bool = False,
        timing: Optional[Union[BitTiming, BitTimingFd]] = None,
        **kwargs: Any,
    ) -> None:
        """
        :param int channel:
            Channel number (zero indexed, labeled on multi-channel devices)
        :param int bitrate:
            Bitrate in bits/s
        :param bool monitor:
            If true, operate in listen-only monitoring mode
        :param timing:
            Optional :class:`~can.BitTiming` instance to use for custom bit timing setting.
            If this argument is set then it overrides the bitrate argument. The
            `f_clock` value of the timing instance must be set to 24_000_000 (24MHz)
            for standard CAN.
            CAN FD and the :class:`~can.BitTimingFd` class are not supported.
        """

        if kwargs.get("_testing", False):
            self.interface = MockInterface()
        else:
            if cantact is None:
                raise CanInterfaceNotImplementedError(
                    "The CANtact module is not installed. "
                    "Install it using `python -m pip install cantact`"
                )
            with error_check(
                "Cannot create the cantact.Interface", CanInitializationError
            ):
                self.interface = cantact.Interface()

        self.channel = int(channel)
        self.channel_info = f"CANtact: ch:{channel}"

        # Configure the interface
        with error_check("Cannot setup the cantact.Interface", CanInitializationError):
            if isinstance(timing, BitTiming):
                timing = check_or_adjust_timing_clock(timing, valid_clocks=[24_000_000])

                # use custom bit timing
                self.interface.set_bit_timing(
                    int(channel),
                    int(timing.brp),
                    int(timing.tseg1),
                    int(timing.tseg2),
                    int(timing.sjw),
                )
            elif isinstance(timing, BitTimingFd):
                raise NotImplementedError(
                    f"CAN FD is not supported by {self.__class__.__name__}."
                )
            else:
                # use bitrate
                self.interface.set_bitrate(int(channel), int(bitrate))

            self.interface.set_enabled(int(channel), True)
            self.interface.set_monitor(int(channel), monitor)
            self.interface.start()

        super().__init__(
            channel=channel, bitrate=bitrate, poll_interval=poll_interval, **kwargs
        )

    def _recv_internal(self, timeout):
        with error_check("Cannot receive message"):
            frame = self.interface.recv(int(timeout * 1000))
        if frame is None:
            # timeout occurred
            return None, False

        msg = Message(
            arbitration_id=frame["id"],
            is_extended_id=frame["extended"],
            timestamp=frame["timestamp"],
            is_remote_frame=frame["rtr"],
            dlc=frame["dlc"],
            data=frame["data"][: frame["dlc"]],
            channel=frame["channel"],
            is_rx=(not frame["loopback"]),  # received if not loopback frame
        )
        return msg, False

    def send(self, msg, timeout=None):
        with error_check("Cannot send message"):
            self.interface.send(
                self.channel,
                msg.arbitration_id,
                bool(msg.is_extended_id),
                bool(msg.is_remote_frame),
                msg.dlc,
                msg.data,
            )

    def shutdown(self):
        super().shutdown()
        with error_check("Cannot shutdown interface"):
            self.interface.stop()


def mock_recv(timeout):
    if timeout > 0:
        return {
            "id": 0x123,
            "extended": False,
            "timestamp": time.time(),
            "loopback": False,
            "rtr": False,
            "dlc": 8,
            "data": [1, 2, 3, 4, 5, 6, 7, 8],
            "channel": 0,
        }
    else:
        # simulate timeout when timeout = 0
        return None


class MockInterface:
    """
    Mock interface to replace real interface when testing.
    This allows for tests to run without actual hardware.
    """

    start = Mock()
    set_bitrate = Mock()
    set_bit_timing = Mock()
    set_enabled = Mock()
    set_monitor = Mock()
    stop = Mock()
    send = Mock()
    channel_count = Mock(return_value=1)

    recv = Mock(side_effect=mock_recv)



"""
Contains the ABC bus implementation and its documentation.
"""

import contextlib
import logging
import threading
from abc import ABC, ABCMeta, abstractmethod
from enum import Enum, auto
from time import time
from typing import Any, Iterator, List, Optional, Sequence, Tuple, Union, cast

import can
import can.typechecking
from can.broadcastmanager import CyclicSendTaskABC, ThreadBasedCyclicSendTask
from can.message import Message

LOG = logging.getLogger(__name__)


class BusState(Enum):
    """The state in which a :class:`can.BusABC` can be."""

    ACTIVE = auto()
    PASSIVE = auto()
    ERROR = auto()


class BusABC(metaclass=ABCMeta):
    """The CAN Bus Abstract Base Class that serves as the basis
    for all concrete interfaces.

    This class may be used as an iterator over the received messages
    and as a context manager for auto-closing the bus when done using it.

    Please refer to :ref:`errors` for possible exceptions that may be
    thrown by certain operations on this bus.
    """

    #: a string describing the underlying bus and/or channel
    channel_info = "unknown"

    #: Log level for received messages
    RECV_LOGGING_LEVEL = 9

    _is_shutdown: bool = False

    @abstractmethod
    def __init__(
        self,
        channel: Any,
        can_filters: Optional[can.typechecking.CanFilters] = None,
        **kwargs: object,
    ):
        """Construct and open a CAN bus instance of the specified type.

        Subclasses should call though this method with all given parameters
        as it handles generic tasks like applying filters.

        :param channel:
            The can interface identifier. Expected type is backend dependent.

        :param can_filters:
            See :meth:`~can.BusABC.set_filters` for details.

        :param dict kwargs:
            Any backend dependent configurations are passed in this dictionary

        :raises ValueError: If parameters are out of range
        :raises ~can.exceptions.CanInterfaceNotImplementedError:
            If the driver cannot be accessed
        :raises ~can.exceptions.CanInitializationError:
            If the bus cannot be initialized
        """
        self._periodic_tasks: List[_SelfRemovingCyclicTask] = []
        self.set_filters(can_filters)

    def __str__(self) -> str:
        return self.channel_info

    def recv(self, timeout: Optional[float] = None) -> Optional[Message]:
        """Block waiting for a message from the Bus.

        :param timeout:
            seconds to wait for a message or None to wait indefinitely

        :return:
            :obj:`None` on timeout or a :class:`~can.Message` object.

        :raises ~can.exceptions.CanOperationError:
            If an error occurred while reading
        """
        start = time()
        time_left = timeout

        while True:
            # try to get a message
            msg, already_filtered = self._recv_internal(timeout=time_left)

            # return it, if it matches
            if msg and (already_filtered or self._matches_filters(msg)):
                LOG.log(self.RECV_LOGGING_LEVEL, "Received: %s", msg)
                return msg

            # if not, and timeout is None, try indefinitely
            elif timeout is None:
                continue

            # try next one only if there still is time, and with
            # reduced timeout
            else:
                time_left = timeout - (time() - start)

                if time_left > 0:
                    continue

                return None

    def _recv_internal(
        self, timeout: Optional[float]
    ) -> Tuple[Optional[Message], bool]:
        """
        Read a message from the bus and tell whether it was filtered.
        This methods may be called by :meth:`~can.BusABC.recv`
        to read a message multiple times if the filters set by
        :meth:`~can.BusABC.set_filters` do not match and the call has
        not yet timed out.

        New implementations should always override this method instead of
        :meth:`~can.BusABC.recv`, to be able to take advantage of the
        software based filtering provided by :meth:`~can.BusABC.recv`
        as a fallback. This method should never be called directly.

        .. note::

            This method is not an `@abstractmethod` (for now) to allow older
            external implementations to continue using their existing
            :meth:`~can.BusABC.recv` implementation.

        .. note::

            The second return value (whether filtering was already done) may
            change over time for some interfaces, like for example in the
            Kvaser interface. Thus it cannot be simplified to a constant value.

        :param float timeout: seconds to wait for a message,
                              see :meth:`~can.BusABC.send`

        :return:
            1.  a message that was read or None on timeout
            2.  a bool that is True if message filtering has already
                been done and else False

        :raises ~can.exceptions.CanOperationError:
            If an error occurred while reading
        :raises NotImplementedError:
            if the bus provides it's own :meth:`~can.BusABC.recv`
            implementation (legacy implementation)

        """
        raise NotImplementedError("Trying to read from a write only bus?")

    @abstractmethod
    def send(self, msg: Message, timeout: Optional[float] = None) -> None:
        """Transmit a message to the CAN bus.

        Override this method to enable the transmit path.

        :param Message msg: A message object.

        :param timeout:
            If > 0, wait up to this many seconds for message to be ACK'ed or
            for transmit queue to be ready depending on driver implementation.
            If timeout is exceeded, an exception will be raised.
            Might not be supported by all interfaces.
            None blocks indefinitely.

        :raises ~can.exceptions.CanOperationError:
            If an error occurred while sending
        """
        raise NotImplementedError("Trying to write to a readonly bus?")

    def send_periodic(
        self,
        msgs: Union[Message, Sequence[Message]],
        period: float,
        duration: Optional[float] = None,
        store_task: bool = True,
    ) -> can.broadcastmanager.CyclicSendTaskABC:
        """Start sending messages at a given period on this bus.

        The task will be active until one of the following conditions are met:

        - the (optional) duration expires
        - the Bus instance goes out of scope
        - the Bus instance is shutdown
        - :meth:`stop_all_periodic_tasks` is called
        - the task's :meth:`~can.broadcastmanager.CyclicTask.stop` method is called.

        :param msgs:
            Message(s) to transmit
        :param period:
            Period in seconds between each message
        :param duration:
            Approximate duration in seconds to continue sending messages. If
            no duration is provided, the task will continue indefinitely.
        :param store_task:
            If True (the default) the task will be attached to this Bus instance.
            Disable to instead manage tasks manually.
        :return:
            A started task instance. Note the task can be stopped (and depending on
            the backend modified) by calling the task's
            :meth:`~can.broadcastmanager.CyclicTask.stop` method.

        .. note::

            Note the duration before the messages stop being sent may not
            be exactly the same as the duration specified by the user. In
            general the message will be sent at the given rate until at
            least **duration** seconds.

        .. note::

            For extremely long running Bus instances with many short lived
            tasks the default api with ``store_task==True`` may not be
            appropriate as the stopped tasks are still taking up memory as they
            are associated with the Bus instance.
        """
        if isinstance(msgs, Message):
            msgs = [msgs]
        elif isinstance(msgs, Sequence):
            # A Sequence does not necessarily provide __bool__ we need to use len()
            if len(msgs) == 0:
                raise ValueError("Must be a sequence at least of length 1")
        else:
            raise ValueError("Must be either a message or a sequence of messages")

        # Create a backend specific task; will be patched to a _SelfRemovingCyclicTask later
        task = cast(
            _SelfRemovingCyclicTask,
            self._send_periodic_internal(msgs, period, duration),
        )

        # we wrap the task's stop method to also remove it from the Bus's list of tasks
        periodic_tasks = self._periodic_tasks
        original_stop_method = task.stop

        def wrapped_stop_method(remove_task: bool = True) -> None:
            nonlocal task, periodic_tasks, original_stop_method
            if remove_task:
                try:
                    periodic_tasks.remove(task)
                except ValueError:
                    pass  # allow the task to be already removed
            original_stop_method()

        task.stop = wrapped_stop_method  # type: ignore

        if store_task:
            self._periodic_tasks.append(task)

        return task

    def _send_periodic_internal(
        self,
        msgs: Union[Sequence[Message], Message],
        period: float,
        duration: Optional[float] = None,
    ) -> can.broadcastmanager.CyclicSendTaskABC:
        """Default implementation of periodic message sending using threading.

        Override this method to enable a more efficient backend specific approach.

        :param msgs:
            Messages to transmit
        :param period:
            Period in seconds between each message
        :param duration:
            The duration between sending each message at the given rate. If
            no duration is provided, the task will continue indefinitely.
        :return:
            A started task instance. Note the task can be stopped (and
            depending on the backend modified) by calling the
            :meth:`~can.broadcastmanager.CyclicTask.stop` method.
        """
        if not hasattr(self, "_lock_send_periodic"):
            # Create a send lock for this bus, but not for buses which override this method
            self._lock_send_periodic = (  # pylint: disable=attribute-defined-outside-init
                threading.Lock()
            )
        task = ThreadBasedCyclicSendTask(
            self, self._lock_send_periodic, msgs, period, duration
        )
        return task

    def stop_all_periodic_tasks(self, remove_tasks: bool = True) -> None:
        """Stop sending any messages that were started using :meth:`send_periodic`.

        .. note::
            The result is undefined if a single task throws an exception while being stopped.

        :param remove_tasks:
            Stop tracking the stopped tasks.
        """
        if not hasattr(self, "_periodic_tasks"):
            # avoid AttributeError for partially initialized BusABC instance
            return

        for task in self._periodic_tasks:
            # we cannot let `task.stop()` modify `self._periodic_tasks` while we are
            # iterating over it (#634)
            task.stop(remove_task=False)

        if remove_tasks:
            self._periodic_tasks.clear()

    def __iter__(self) -> Iterator[Message]:
        """Allow iteration on messages as they are received.

        .. code-block:: python

            for msg in bus:
                print(msg)


        :yields:
            :class:`Message` msg objects.
        """
        while True:
            msg = self.recv(timeout=1.0)
            if msg is not None:
                yield msg

    @property
    def filters(self) -> Optional[can.typechecking.CanFilters]:
        """
        Modify the filters of this bus. See :meth:`~can.BusABC.set_filters`
        for details.
        """
        return self._filters

    @filters.setter
    def filters(self, filters: Optional[can.typechecking.CanFilters]) -> None:
        self.set_filters(filters)

    def set_filters(
        self, filters: Optional[can.typechecking.CanFilters] = None
    ) -> None:
        """Apply filtering to all messages received by this Bus.

        All messages that match at least one filter are returned.
        If `filters` is `None` or a zero length sequence, all
        messages are matched.

        Calling without passing any filters will reset the applied
        filters to ``None``.

        :param filters:
            A iterable of dictionaries each containing a "can_id",
            a "can_mask", and an optional "extended" key::

                [{"can_id": 0x11, "can_mask": 0x21, "extended": False}]

            A filter matches, when
            ``<received_can_id> & can_mask == can_id & can_mask``.
            If ``extended`` is set as well, it only matches messages where
            ``<received_is_extended> == extended``. Else it matches every
            messages based only on the arbitration ID and mask.
        """
        self._filters = filters or None
        self._apply_filters(self._filters)

    def _apply_filters(self, filters: Optional[can.typechecking.CanFilters]) -> None:
        """
        Hook for applying the filters to the underlying kernel or
        hardware if supported/implemented by the interface.

        :param filters:
            See :meth:`~can.BusABC.set_filters` for details.
        """

    def _matches_filters(self, msg: Message) -> bool:
        """Checks whether the given message matches at least one of the
        current filters. See :meth:`~can.BusABC.set_filters` for details
        on how the filters work.

        This method should not be overridden.

        :param msg:
            the message to check if matching
        :return: whether the given message matches at least one filter
        """

        # if no filters are set, all messages are matched
        if self._filters is None:
            return True

        for _filter in self._filters:
            # check if this filter even applies to the message
            if "extended" in _filter:
                _filter = cast(can.typechecking.CanFilterExtended, _filter)
                if _filter["extended"] != msg.is_extended_id:
                    continue

            # then check for the mask and id
            can_id = _filter["can_id"]
            can_mask = _filter["can_mask"]

            # basically, we compute
            # `msg.arbitration_id & can_mask == can_id & can_mask`
            # by using the shorter, but equivalent from below:
            if (can_id ^ msg.arbitration_id) & can_mask == 0:
                return True

        # nothing matched
        return False

    def flush_tx_buffer(self) -> None:
        """Discard every message that may be queued in the output buffer(s)."""

    def shutdown(self) -> None:
        """
        Called to carry out any interface specific cleanup required in shutting down a bus.

        This method can be safely called multiple times.
        """
        if self._is_shutdown:
            LOG.debug("%s is already shut down", self.__class__)
            return

        self._is_shutdown = True
        self.stop_all_periodic_tasks()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()

    def __del__(self) -> None:
        if not self._is_shutdown:
            LOG.warning("%s was not properly shut down", self.__class__)
            # We do some best-effort cleanup if the user
            # forgot to properly close the bus instance
            with contextlib.suppress(AttributeError):
                self.shutdown()

    @property
    def state(self) -> BusState:
        """
        Return the current state of the hardware
        """
        return BusState.ACTIVE

    @state.setter
    def state(self, new_state: BusState) -> None:
        """
        Set the new state of the hardware
        """
        raise NotImplementedError("Property is not implemented.")

    @staticmethod
    def _detect_available_configs() -> List[can.typechecking.AutoDetectedConfig]:
        """Detect all configurations/channels that this interface could
        currently connect with.

        This might be quite time consuming.

        May not to be implemented by every interface on every platform.

        :return: an iterable of dicts, each being a configuration suitable
                 for usage in the interface's bus constructor.
        """
        raise NotImplementedError()

    def fileno(self) -> int:
        raise NotImplementedError("fileno is not implemented using current CAN bus")


class _SelfRemovingCyclicTask(CyclicSendTaskABC, ABC):
    """Removes itself from a bus.

    Only needed for typing :meth:`Bus._periodic_tasks`. Do not instantiate.
    """

    def stop(self, remove_task: bool = True) -> None:
        raise NotImplementedError()



#!/usr/bin/env python

"""
"""
import asyncio
import logging
import os
import random
import tempfile
import unittest
import warnings
from os.path import dirname, join

import can

from .data.example_data import generate_message

logging.basicConfig(level=logging.DEBUG)

# makes the random number generator deterministic
random.seed(13339115)


class ListenerImportTest(unittest.TestCase):
    def testClassesImportable(self):
        self.assertTrue(hasattr(can, "Listener"))
        self.assertTrue(hasattr(can, "BufferedReader"))
        self.assertTrue(hasattr(can, "Notifier"))
        self.assertTrue(hasattr(can, "Logger"))

        self.assertTrue(hasattr(can, "ASCWriter"))
        self.assertTrue(hasattr(can, "ASCReader"))

        self.assertTrue(hasattr(can, "BLFReader"))
        self.assertTrue(hasattr(can, "BLFWriter"))

        self.assertTrue(hasattr(can, "CSVReader"))
        self.assertTrue(hasattr(can, "CSVWriter"))

        self.assertTrue(hasattr(can, "CanutilsLogWriter"))
        self.assertTrue(hasattr(can, "CanutilsLogReader"))

        self.assertTrue(hasattr(can, "SqliteReader"))
        self.assertTrue(hasattr(can, "SqliteWriter"))

        self.assertTrue(hasattr(can, "Printer"))

        self.assertTrue(hasattr(can, "LogReader"))

        self.assertTrue(hasattr(can, "MessageSync"))


class BusTest(unittest.TestCase):
    def setUp(self):
        # Save all can.rc defaults
        self._can_rc = can.rc
        can.rc = {"interface": "virtual"}
        self.bus = can.interface.Bus()

    def tearDown(self):
        self.bus.shutdown()
        # Restore the defaults
        can.rc = self._can_rc


class ListenerTest(BusTest):
    def testBasicListenerCanBeAddedToNotifier(self):
        a_listener = can.Printer()
        notifier = can.Notifier(self.bus, [a_listener], 0.1)
        notifier.stop()
        self.assertIn(a_listener, notifier.listeners)

    def testAddListenerToNotifier(self):
        a_listener = can.Printer()
        notifier = can.Notifier(self.bus, [], 0.1)
        notifier.stop()
        self.assertNotIn(a_listener, notifier.listeners)
        notifier.add_listener(a_listener)
        self.assertIn(a_listener, notifier.listeners)

    def testRemoveListenerFromNotifier(self):
        a_listener = can.Printer()
        notifier = can.Notifier(self.bus, [a_listener], 0.1)
        notifier.stop()
        self.assertIn(a_listener, notifier.listeners)
        notifier.remove_listener(a_listener)
        self.assertNotIn(a_listener, notifier.listeners)

    def testPlayerTypeResolution(self):
        def test_filetype_to_instance(extension, klass):
            print(f"testing: {extension}")
            try:
                if extension == ".blf":
                    delete = False
                    file_handler = open(
                        join(dirname(__file__), "data", "test_CanMessage.blf")
                    )
                else:
                    delete = True
                    file_handler = tempfile.NamedTemporaryFile(
                        suffix=extension, delete=False
                    )

                with file_handler as my_file:
                    filename = my_file.name
                with can.LogReader(filename) as reader:
                    self.assertIsInstance(reader, klass)
            finally:
                if delete:
                    os.remove(filename)

        test_filetype_to_instance(".asc", can.ASCReader)
        test_filetype_to_instance(".blf", can.BLFReader)
        test_filetype_to_instance(".csv", can.CSVReader)
        test_filetype_to_instance(".db", can.SqliteReader)
        test_filetype_to_instance(".log", can.CanutilsLogReader)

    def testPlayerTypeResolutionUnsupportedFileTypes(self):
        for should_fail_with in ["", ".", ".some_unknown_extention_42"]:
            with self.assertRaises(ValueError):
                with can.LogReader(should_fail_with):  # make sure we close it anyways
                    pass

    def testLoggerTypeResolution(self):
        def test_filetype_to_instance(extension, klass):
            print(f"testing: {extension}")
            try:
                with tempfile.NamedTemporaryFile(
                    suffix=extension, delete=False
                ) as my_file:
                    filename = my_file.name
                with can.Logger(filename) as writer:
                    self.assertIsInstance(writer, klass)
            finally:
                os.remove(filename)

        test_filetype_to_instance(".asc", can.ASCWriter)
        test_filetype_to_instance(".blf", can.BLFWriter)
        test_filetype_to_instance(".csv", can.CSVWriter)
        test_filetype_to_instance(".db", can.SqliteWriter)
        test_filetype_to_instance(".log", can.CanutilsLogWriter)
        test_filetype_to_instance(".txt", can.Printer)

        with can.Logger(None) as logger:
            self.assertIsInstance(logger, can.Printer)

    def testLoggerTypeResolutionUnsupportedFileTypes(self):
        for should_fail_with in ["", ".", ".some_unknown_extention_42"]:
            with self.assertRaises(ValueError):
                with can.Logger(should_fail_with):  # make sure we close it anyways
                    pass

    def testBufferedListenerReceives(self):
        a_listener = can.BufferedReader()
        a_listener(generate_message(0xDADADA))
        a_listener(generate_message(0xDADADA))
        self.assertIsNotNone(a_listener.get_message(0.1))
        a_listener.stop()
        self.assertIsNotNone(a_listener.get_message(0.1))


def test_deprecated_loop_arg(recwarn):
    warnings.simplefilter("always")
    can.AsyncBufferedReader(loop=asyncio.get_event_loop())
    assert len(recwarn) > 0
    assert recwarn.pop(DeprecationWarning)
    recwarn.clear()

    # assert that no warning is shown when loop argument is not used
    can.AsyncBufferedReader()
    assert len(recwarn) == 0


if __name__ == "__main__":
    unittest.main()



"""
Contains Python equivalents of the structures in CANLIB's canlib.h,
with some supporting functionality specific to Python.
"""

import ctypes


class BusStatistics(ctypes.Structure):
    """This structure is used with the method
    :meth:`~can.interfaces.kvaser.canlib.KvaserBus.get_stats`.
    """

    _fields_ = [
        ("m_stdData", ctypes.c_ulong),
        ("m_stdRemote", ctypes.c_ulong),
        ("m_extData", ctypes.c_ulong),
        ("m_extRemote", ctypes.c_ulong),
        ("m_errFrame", ctypes.c_ulong),
        ("m_busLoad", ctypes.c_ulong),
        ("m_overruns", ctypes.c_ulong),
    ]

    def __str__(self):
        return (
            "std_data: {}, std_remote: {}, ext_data: {}, ext_remote: {}, "
            "err_frame: {}, bus_load: {:.1f}%, overruns: {}"
        ).format(
            self.std_data,
            self.std_remote,
            self.ext_data,
            self.ext_remote,
            self.err_frame,
            self.bus_load / 100.0,
            self.overruns,
        )

    @property
    def std_data(self):
        """Number of received standard (11-bit identifiers) data frames."""
        return self.m_stdData

    @property
    def std_remote(self):
        """Number of received standard (11-bit identifiers) remote frames."""
        return self.m_stdRemote

    @property
    def ext_data(self):
        """Number of received extended (29-bit identifiers) data frames."""
        return self.m_extData

    @property
    def ext_remote(self):
        """Number of received extended (29-bit identifiers) remote frames."""
        return self.m_extRemote

    @property
    def err_frame(self):
        """Number of error frames."""
        return self.m_errFrame

    @property
    def bus_load(self):
        """The bus load, expressed as an integer in the interval 0 - 10000 representing 0.00% - 100.00% bus load."""
        return self.m_busLoad

    @property
    def overruns(self):
        """Number of overruns."""
        return self.m_overruns



"""
There are several specific :class:`Exception` classes to allow user
code to react to specific scenarios related to CAN busses::

    Exception (Python standard library)
     +-- ...
     +-- CanError (python-can)
         +-- CanInterfaceNotImplementedError
         +-- CanInitializationError
         +-- CanOperationError
         +-- CanTimeoutError

Keep in mind that some functions and methods may raise different exceptions.
For example, validating typical arguments and parameters might result in a
:class:`ValueError`. This should always be documented for the function at hand.
"""

import sys
from contextlib import contextmanager
from typing import Optional, Type

if sys.version_info >= (3, 9):
    from collections.abc import Generator
else:
    from typing import Generator


class CanError(Exception):
    """Base class for all CAN related exceptions.

    If specified, the error code is automatically appended to the message:

    .. testsetup:: canerror

        from can import CanError, CanOperationError

    .. doctest:: canerror

        >>> # With an error code (it also works with a specific error):
        >>> error = CanOperationError(message="Failed to do the thing", error_code=42)
        >>> str(error)
        'Failed to do the thing [Error Code 42]'
        >>>
        >>> # Missing the error code:
        >>> plain_error = CanError(message="Something went wrong ...")
        >>> str(plain_error)
        'Something went wrong ...'

    :param error_code:
        An optional error code to narrow down the cause of the fault

    :arg error_code:
        An optional error code to narrow down the cause of the fault
    """

    def __init__(
        self,
        message: str = "",
        error_code: Optional[int] = None,
    ) -> None:
        self.error_code = error_code
        super().__init__(
            message if error_code is None else f"{message} [Error Code {error_code}]"
        )


class CanInterfaceNotImplementedError(CanError, NotImplementedError):
    """Indicates that the interface is not supported on the current platform.

    Example scenarios:
      - No interface with that name exists
      - The interface is unsupported on the current operating system or interpreter
      - The driver could not be found or has the wrong version
    """


class CanInitializationError(CanError):
    """Indicates an error the occurred while initializing a :class:`can.BusABC`.

    If initialization fails due to a driver or platform missing/being unsupported,
    a :exc:`~can.exceptions.CanInterfaceNotImplementedError` is raised instead.
    If initialization fails due to a value being out of range, a :class:`ValueError`
    is raised.

    Example scenarios:
      - Try to open a non-existent device and/or channel
      - Try to use an invalid setting, which is ok by value, but not ok for the interface
      - The device or other resources are already used
    """


class CanOperationError(CanError):
    """Indicates an error while in operation.

    Example scenarios:
      - A call to a library function results in an unexpected return value
      - An invalid message was received
      - The driver rejected a message that was meant to be sent
      - Cyclic redundancy check (CRC) failed
      - A message remained unacknowledged
      - A buffer is full
    """


class CanTimeoutError(CanError, TimeoutError):
    """Indicates the timeout of an operation.

    Example scenarios:
      - Some message could not be sent after the timeout elapsed
      - No message was read within the given time
    """


@contextmanager
def error_check(
    error_message: Optional[str] = None,
    exception_type: Type[CanError] = CanOperationError,
) -> Generator[None, None, None]:
    """Catches any exceptions and turns them into the new type while preserving the stack trace."""
    try:
        yield
    except Exception as error:  # pylint: disable=broad-except
        if error_message is None:
            raise exception_type(str(error)) from error
        else:
            raise exception_type(error_message) from error



# type: ignore
"""
Ctypes wrapper module for Vector CAN Interface on win32/win64 systems.

Authors: Julien Grave <grave.jul@gmail.com>, Christian Sandberg
"""

# Import Standard Python Modules
# ==============================
import ctypes
import logging
import platform

from .exceptions import VectorInitializationError, VectorOperationError

# Define Module Logger
# ====================
LOG = logging.getLogger(__name__)

# Vector XL API Definitions
# =========================
from . import xlclass

# Load Windows DLL
DLL_NAME = "vxlapi64" if platform.architecture()[0] == "64bit" else "vxlapi"
_xlapi_dll = ctypes.windll.LoadLibrary(DLL_NAME)


# ctypes wrapping for API functions
xlGetErrorString = _xlapi_dll.xlGetErrorString
xlGetErrorString.argtypes = [xlclass.XLstatus]
xlGetErrorString.restype = xlclass.XLstringType


def check_status_operation(result, function, arguments):
    """Check the status and raise a :class:`VectorOperationError` on error."""
    if result > 0:
        raise VectorOperationError(
            result, xlGetErrorString(result).decode(), function.__name__
        )
    return result


def check_status_initialization(result, function, arguments):
    """Check the status and raise a :class:`VectorInitializationError` on error."""
    if result > 0:
        raise VectorInitializationError(
            result, xlGetErrorString(result).decode(), function.__name__
        )
    return result


xlGetDriverConfig = _xlapi_dll.xlGetDriverConfig
xlGetDriverConfig.argtypes = [ctypes.POINTER(xlclass.XLdriverConfig)]
xlGetDriverConfig.restype = xlclass.XLstatus
xlGetDriverConfig.errcheck = check_status_operation

xlOpenDriver = _xlapi_dll.xlOpenDriver
xlOpenDriver.argtypes = []
xlOpenDriver.restype = xlclass.XLstatus
xlOpenDriver.errcheck = check_status_initialization

xlCloseDriver = _xlapi_dll.xlCloseDriver
xlCloseDriver.argtypes = []
xlCloseDriver.restype = xlclass.XLstatus
xlCloseDriver.errcheck = check_status_operation

xlGetApplConfig = _xlapi_dll.xlGetApplConfig
xlGetApplConfig.argtypes = [
    ctypes.c_char_p,
    ctypes.c_uint,
    ctypes.POINTER(ctypes.c_uint),
    ctypes.POINTER(ctypes.c_uint),
    ctypes.POINTER(ctypes.c_uint),
    ctypes.c_uint,
]
xlGetApplConfig.restype = xlclass.XLstatus
xlGetApplConfig.errcheck = check_status_initialization

xlSetApplConfig = _xlapi_dll.xlSetApplConfig
xlSetApplConfig.argtypes = [
    ctypes.c_char_p,
    ctypes.c_uint,
    ctypes.c_uint,
    ctypes.c_uint,
    ctypes.c_uint,
    ctypes.c_uint,
]
xlSetApplConfig.restype = xlclass.XLstatus
xlSetApplConfig.errcheck = check_status_initialization

xlGetChannelIndex = _xlapi_dll.xlGetChannelIndex
xlGetChannelIndex.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_int]
xlGetChannelIndex.restype = ctypes.c_int

xlGetChannelMask = _xlapi_dll.xlGetChannelMask
xlGetChannelMask.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_int]
xlGetChannelMask.restype = xlclass.XLaccess

xlOpenPort = _xlapi_dll.xlOpenPort
xlOpenPort.argtypes = [
    ctypes.POINTER(xlclass.XLportHandle),
    ctypes.c_char_p,
    xlclass.XLaccess,
    ctypes.POINTER(xlclass.XLaccess),
    ctypes.c_uint,
    ctypes.c_uint,
    ctypes.c_uint,
]
xlOpenPort.restype = xlclass.XLstatus
xlOpenPort.errcheck = check_status_initialization

xlGetSyncTime = _xlapi_dll.xlGetSyncTime
xlGetSyncTime.argtypes = [xlclass.XLportHandle, ctypes.POINTER(xlclass.XLuint64)]
xlGetSyncTime.restype = xlclass.XLstatus
xlGetSyncTime.errcheck = check_status_initialization

xlGetChannelTime = _xlapi_dll.xlGetChannelTime
xlGetChannelTime.argtypes = [
    xlclass.XLportHandle,
    xlclass.XLaccess,
    ctypes.POINTER(xlclass.XLuint64),
]
xlGetChannelTime.restype = xlclass.XLstatus
xlGetChannelTime.errcheck = check_status_initialization

xlClosePort = _xlapi_dll.xlClosePort
xlClosePort.argtypes = [xlclass.XLportHandle]
xlClosePort.restype = xlclass.XLstatus
xlClosePort.errcheck = check_status_operation

xlSetNotification = _xlapi_dll.xlSetNotification
xlSetNotification.argtypes = [
    xlclass.XLportHandle,
    ctypes.POINTER(xlclass.XLhandle),
    ctypes.c_int,
]
xlSetNotification.restype = xlclass.XLstatus
xlSetNotification.errcheck = check_status_initialization

xlCanSetChannelMode = _xlapi_dll.xlCanSetChannelMode
xlCanSetChannelMode.argtypes = [
    xlclass.XLportHandle,
    xlclass.XLaccess,
    ctypes.c_int,
    ctypes.c_int,
]
xlCanSetChannelMode.restype = xlclass.XLstatus
xlCanSetChannelMode.errcheck = check_status_initialization

xlActivateChannel = _xlapi_dll.xlActivateChannel
xlActivateChannel.argtypes = [
    xlclass.XLportHandle,
    xlclass.XLaccess,
    ctypes.c_uint,
    ctypes.c_uint,
]
xlActivateChannel.restype = xlclass.XLstatus
xlActivateChannel.errcheck = check_status_operation

xlDeactivateChannel = _xlapi_dll.xlDeactivateChannel
xlDeactivateChannel.argtypes = [xlclass.XLportHandle, xlclass.XLaccess]
xlDeactivateChannel.restype = xlclass.XLstatus
xlDeactivateChannel.errcheck = check_status_operation

xlCanFdSetConfiguration = _xlapi_dll.xlCanFdSetConfiguration
xlCanFdSetConfiguration.argtypes = [
    xlclass.XLportHandle,
    xlclass.XLaccess,
    ctypes.POINTER(xlclass.XLcanFdConf),
]
xlCanFdSetConfiguration.restype = xlclass.XLstatus
xlCanFdSetConfiguration.errcheck = check_status_initialization

xlReceive = _xlapi_dll.xlReceive
xlReceive.argtypes = [
    xlclass.XLportHandle,
    ctypes.POINTER(ctypes.c_uint),
    ctypes.POINTER(xlclass.XLevent),
]
xlReceive.restype = xlclass.XLstatus
xlReceive.errcheck = check_status_operation

xlCanReceive = _xlapi_dll.xlCanReceive
xlCanReceive.argtypes = [xlclass.XLportHandle, ctypes.POINTER(xlclass.XLcanRxEvent)]
xlCanReceive.restype = xlclass.XLstatus
xlCanReceive.errcheck = check_status_operation

xlCanSetChannelBitrate = _xlapi_dll.xlCanSetChannelBitrate
xlCanSetChannelBitrate.argtypes = [
    xlclass.XLportHandle,
    xlclass.XLaccess,
    ctypes.c_ulong,
]
xlCanSetChannelBitrate.restype = xlclass.XLstatus
xlCanSetChannelBitrate.errcheck = check_status_initialization

xlCanSetChannelParams = _xlapi_dll.xlCanSetChannelParams
xlCanSetChannelParams.argtypes = [
    xlclass.XLportHandle,
    xlclass.XLaccess,
    ctypes.POINTER(xlclass.XLchipParams),
]
xlCanSetChannelParams.restype = xlclass.XLstatus
xlCanSetChannelParams.errcheck = check_status_initialization

xlCanSetChannelParamsC200 = _xlapi_dll.xlCanSetChannelParamsC200
xlCanSetChannelParamsC200.argtypes = [
    xlclass.XLportHandle,
    xlclass.XLaccess,
    ctypes.c_ubyte,
    ctypes.c_ubyte,
]
xlCanSetChannelParams.restype = xlclass.XLstatus
xlCanSetChannelParams.errcheck = check_status_initialization

xlCanTransmit = _xlapi_dll.xlCanTransmit
xlCanTransmit.argtypes = [
    xlclass.XLportHandle,
    xlclass.XLaccess,
    ctypes.POINTER(ctypes.c_uint),
    ctypes.POINTER(xlclass.XLevent),
]
xlCanTransmit.restype = xlclass.XLstatus
xlCanTransmit.errcheck = check_status_operation

xlCanTransmitEx = _xlapi_dll.xlCanTransmitEx
xlCanTransmitEx.argtypes = [
    xlclass.XLportHandle,
    xlclass.XLaccess,
    ctypes.c_uint,
    ctypes.POINTER(ctypes.c_uint),
    ctypes.POINTER(xlclass.XLcanTxEvent),
]
xlCanTransmitEx.restype = xlclass.XLstatus
xlCanTransmitEx.errcheck = check_status_operation

xlCanFlushTransmitQueue = _xlapi_dll.xlCanFlushTransmitQueue
xlCanFlushTransmitQueue.argtypes = [xlclass.XLportHandle, xlclass.XLaccess]
xlCanFlushTransmitQueue.restype = xlclass.XLstatus
xlCanFlushTransmitQueue.errcheck = check_status_operation

xlCanSetChannelAcceptance = _xlapi_dll.xlCanSetChannelAcceptance
xlCanSetChannelAcceptance.argtypes = [
    xlclass.XLportHandle,
    xlclass.XLaccess,
    ctypes.c_ulong,
    ctypes.c_ulong,
    ctypes.c_uint,
]
xlCanSetChannelAcceptance.restype = xlclass.XLstatus
xlCanSetChannelAcceptance.errcheck = check_status_operation

xlCanResetAcceptance = _xlapi_dll.xlCanResetAcceptance
xlCanResetAcceptance.argtypes = [xlclass.XLportHandle, xlclass.XLaccess, ctypes.c_uint]
xlCanResetAcceptance.restype = xlclass.XLstatus
xlCanResetAcceptance.errcheck = check_status_operation

xlCanRequestChipState = _xlapi_dll.xlCanRequestChipState
xlCanRequestChipState.argtypes = [xlclass.XLportHandle, xlclass.XLaccess]
xlCanRequestChipState.restype = xlclass.XLstatus
xlCanRequestChipState.errcheck = check_status_operation

xlCanSetChannelOutput = _xlapi_dll.xlCanSetChannelOutput
xlCanSetChannelOutput.argtypes = [xlclass.XLportHandle, xlclass.XLaccess, ctypes.c_char]
xlCanSetChannelOutput.restype = xlclass.XLstatus
xlCanSetChannelOutput.errcheck = check_status_operation

xlPopupHwConfig = _xlapi_dll.xlPopupHwConfig
xlPopupHwConfig.argtypes = [ctypes.c_char_p, ctypes.c_uint]
xlPopupHwConfig.restype = xlclass.XLstatus
xlPopupHwConfig.errcheck = check_status_operation

xlSetTimerRate = _xlapi_dll.xlSetTimerRate
xlSetTimerRate.argtypes = [xlclass.XLportHandle, ctypes.c_ulong]
xlSetTimerRate.restype = xlclass.XLstatus
xlSetTimerRate.errcheck = check_status_operation

xlGetEventString = _xlapi_dll.xlGetEventString
xlGetEventString.argtypes = [ctypes.POINTER(xlclass.XLevent)]
xlGetEventString.restype = xlclass.XLstringType

xlCanGetEventString = _xlapi_dll.xlCanGetEventString
xlCanGetEventString.argtypes = [ctypes.POINTER(xlclass.XLcanRxEvent)]
xlCanGetEventString.restype = xlclass.XLstringType

xlGetReceiveQueueLevel = _xlapi_dll.xlGetReceiveQueueLevel
xlGetReceiveQueueLevel.argtypes = [xlclass.XLportHandle, ctypes.POINTER(ctypes.c_int)]
xlGetReceiveQueueLevel.restype = xlclass.XLstatus
xlGetReceiveQueueLevel.errcheck = check_status_operation

xlGenerateSyncPulse = _xlapi_dll.xlGenerateSyncPulse
xlGenerateSyncPulse.argtypes = [xlclass.XLportHandle, xlclass.XLaccess]
xlGenerateSyncPulse.restype = xlclass.XLstatus
xlGenerateSyncPulse.errcheck = check_status_operation

xlFlushReceiveQueue = _xlapi_dll.xlFlushReceiveQueue
xlFlushReceiveQueue.argtypes = [xlclass.XLportHandle]
xlFlushReceiveQueue.restype = xlclass.XLstatus
xlFlushReceiveQueue.errcheck = check_status_operation



"""
Ctypes wrapper module for IXXAT Virtual CAN Interface V4 on win32 systems

Copyright (C) 2016 Giuseppe Corbelli <giuseppe.corbelli@weightpack.com>
"""

import ctypes


class LUID(ctypes.Structure):
    _fields_ = [("LowPart", ctypes.c_uint32), ("HighPart", ctypes.c_int32)]


PLUID = ctypes.POINTER(LUID)


class VCIID(ctypes.Union):
    _fields_ = [("AsLuid", LUID), ("AsInt64", ctypes.c_int64)]


PVCIID = ctypes.POINTER(VCIID)


class GUID(ctypes.Structure):
    _fields_ = [
        ("Data1", ctypes.c_uint32),
        ("Data2", ctypes.c_uint16),
        ("Data3", ctypes.c_uint16),
        ("Data4", ctypes.c_char * 8),
    ]


class VCIDEVICEINFO(ctypes.Structure):
    class UniqueHardwareId(ctypes.Union):
        _fields_ = [("AsChar", ctypes.c_char * 16), ("AsGuid", GUID)]

    _fields_ = [
        ("VciObjectId", VCIID),
        ("DeviceClass", GUID),
        ("DriverMajorVersion", ctypes.c_uint8),
        ("DriverMinorVersion", ctypes.c_uint8),
        ("DriverBuildVersion", ctypes.c_uint16),
        ("HardwareBranchVersion", ctypes.c_uint8),
        ("HardwareMajorVersion", ctypes.c_uint8),
        ("HardwareMinorVersion", ctypes.c_uint8),
        ("HardwareBuildVersion", ctypes.c_uint8),
        ("UniqueHardwareId", UniqueHardwareId),
        ("Description", ctypes.c_char * 128),
        ("Manufacturer", ctypes.c_char * 126),
        ("DriverReleaseVersion", ctypes.c_uint16),
    ]

    def __str__(self):
        return "Mfg: {}, Dev: {} HW: {}.{}.{}.{} Drv: {}.{}.{}.{}".format(
            self.Manufacturer,
            self.Description,
            self.HardwareBranchVersion,
            self.HardwareMajorVersion,
            self.HardwareMinorVersion,
            self.HardwareBuildVersion,
            self.DriverReleaseVersion,
            self.DriverMajorVersion,
            self.DriverMinorVersion,
            self.DriverBuildVersion,
        )


PVCIDEVICEINFO = ctypes.POINTER(VCIDEVICEINFO)


class CANLINESTATUS(ctypes.Structure):
    _fields_ = [
        # current CAN operating mode. Value is a logical combination of
        # one or more CAN_OPMODE_xxx constants
        ("bOpMode", ctypes.c_uint8),
        ("bBtReg0", ctypes.c_uint8),  # current bus timing register 0 value
        ("bBtReg1", ctypes.c_uint8),  # current bus timing register 1 value
        ("bBusLoad", ctypes.c_uint8),  # average bus load in percent (0..100)
        ("dwStatus", ctypes.c_uint32),  # status of the CAN controller (see CAN_STATUS_)
    ]


PCANLINESTATUS = ctypes.POINTER(CANLINESTATUS)


class CANCHANSTATUS(ctypes.Structure):
    _fields_ = [
        ("sLineStatus", CANLINESTATUS),  # current CAN line status
        ("fActivated", ctypes.c_uint32),  # TRUE if the channel is activated
        ("fRxOverrun", ctypes.c_uint32),  # TRUE if receive FIFO overrun occurred
        ("bRxFifoLoad", ctypes.c_uint8),  # receive FIFO load in percent (0..100)
        ("bTxFifoLoad", ctypes.c_uint8),  # transmit FIFO load in percent (0..100)
    ]


PCANCHANSTATUS = ctypes.POINTER(CANCHANSTATUS)


class CANCAPABILITIES(ctypes.Structure):
    _fields_ = [
        ("wCtrlType", ctypes.c_uint16),
        ("wBusCoupling", ctypes.c_uint16),
        ("dwFeatures", ctypes.c_uint32),
        ("dwClockFreq", ctypes.c_uint32),
        ("dwTscDivisor", ctypes.c_uint32),
        ("dwCmsDivisor", ctypes.c_uint32),
        ("dwCmsMaxTicks", ctypes.c_uint32),
        ("dwDtxDivisor", ctypes.c_uint32),
        ("dwDtxMaxTicks", ctypes.c_uint32),
    ]


PCANCAPABILITIES = ctypes.POINTER(CANCAPABILITIES)


class CANMSGINFO(ctypes.Union):
    class Bytes(ctypes.Structure):
        _fields_ = [
            ("bType", ctypes.c_uint8),  # type (see CAN_MSGTYPE_ constants)
            (
                "bAddFlags",
                ctypes.c_uint8,
            ),  # extended flags (see CAN_MSGFLAGS2_ constants). AKA bFlags2 in VCI v4
            ("bFlags", ctypes.c_uint8),  # flags (see CAN_MSGFLAGS_ constants)
            ("bAccept", ctypes.c_uint8),  # accept code (see CAN_ACCEPT_ constants)
        ]

    class Bits(ctypes.Structure):
        _fields_ = [
            ("type", ctypes.c_uint32, 8),  # type (see CAN_MSGTYPE_ constants)
            ("ssm", ctypes.c_uint32, 1),  # single shot mode
            ("hpm", ctypes.c_uint32, 1),  # high priority message
            ("edl", ctypes.c_uint32, 1),  # extended data length
            ("fdr", ctypes.c_uint32, 1),  # fast data bit rate
            ("esi", ctypes.c_uint32, 1),  # error state indicator
            ("res", ctypes.c_uint32, 3),  # reserved set to 0
            ("dlc", ctypes.c_uint32, 4),  # data length code
            ("ovr", ctypes.c_uint32, 1),  # data overrun
            ("srr", ctypes.c_uint32, 1),  # self reception request
            ("rtr", ctypes.c_uint32, 1),  # remote transmission request
            (
                "ext",
                ctypes.c_uint32,
                1,
            ),  # extended frame format (0=standard, 1=extended)
            ("afc", ctypes.c_uint32, 8),  # accept code (see CAN_ACCEPT_ constants)
        ]

    _fields_ = [("Bytes", Bytes), ("Bits", Bits)]


PCANMSGINFO = ctypes.POINTER(CANMSGINFO)


class CANMSG(ctypes.Structure):
    _fields_ = [
        ("dwTime", ctypes.c_uint32),
        # CAN ID of the message in Intel format (aligned right) without RTR bit.
        ("dwMsgId", ctypes.c_uint32),
        ("uMsgInfo", CANMSGINFO),
        ("abData", ctypes.c_uint8 * 8),
    ]

    def __str__(self) -> str:
        return """ID: 0x{:04x}{} DLC: {:02d} DATA: {}""".format(
            self.dwMsgId,
            "[RTR]" if self.uMsgInfo.Bits.rtr else "",
            self.uMsgInfo.Bits.dlc,
            memoryview(self.abData)[: self.uMsgInfo.Bits.dlc].hex(sep=" "),
        )


PCANMSG = ctypes.POINTER(CANMSG)


class CANCYCLICTXMSG(ctypes.Structure):
    _fields_ = [
        ("wCycleTime", ctypes.c_uint16),
        ("bIncrMode", ctypes.c_uint8),
        ("bByteIndex", ctypes.c_uint8),
        ("dwMsgId", ctypes.c_uint32),
        ("uMsgInfo", CANMSGINFO),
        ("abData", ctypes.c_uint8 * 8),
    ]


PCANCYCLICTXMSG = ctypes.POINTER(CANCYCLICTXMSG)


class CANBTP(ctypes.Structure):
    _fields_ = [
        ("dwMode", ctypes.c_uint32),  # timing mode (see CAN_BTMODE_ const)
        ("dwBPS", ctypes.c_uint32),  # bits per second or prescaler (see CAN_BTMODE_RAW)
        ("wTS1", ctypes.c_uint16),  # length of time segment 1 in quanta
        ("wTS2", ctypes.c_uint16),  # length of time segment 2 in quanta
        ("wSJW", ctypes.c_uint16),  # re-synchronization jump width im quanta
        (
            "wTDO",
            ctypes.c_uint16,
        ),  # transceiver delay offset (SSP offset) in quanta (0 = disabled, 0xFFFF = simplified SSP positioning)
    ]

    def __str__(self):
        return "dwMode=%d, dwBPS=%d, wTS1=%d,  wTS2=%d, wSJW=%d, wTDO=%d" % (
            self.dwMode,
            self.dwBPS,
            self.wTS1,
            self.wTS2,
            self.wSJW,
            self.wTDO,
        )


PCANBTP = ctypes.POINTER(CANBTP)


class CANCAPABILITIES2(ctypes.Structure):
    _fields_ = [
        ("wCtrlType", ctypes.c_uint16),  # Type of CAN controller (see CAN_CTRL_ const)
        ("wBusCoupling", ctypes.c_uint16),  # Type of Bus coupling (see CAN_BUSC_ const)
        (
            "dwFeatures",
            ctypes.c_uint32,
        ),  # supported features (see CAN_FEATURE_ constants)
        ("dwCanClkFreq", ctypes.c_uint32),  # CAN clock frequency [Hz]
        ("sSdrRangeMin", CANBTP),  # minimum bit timing values for standard bit rate
        ("sSdrRangeMax", CANBTP),  # maximum bit timing values for standard bit rate
        ("sFdrRangeMin", CANBTP),  # minimum bit timing values for fast data bit rate
        ("sFdrRangeMax", CANBTP),  # maximum bit timing values for fast data bit rate
        (
            "dwTscClkFreq",
            ctypes.c_uint32,
        ),  # clock frequency of the time stamp counter [Hz]
        ("dwTscDivisor", ctypes.c_uint32),  # divisor for the message time stamp counter
        (
            "dwCmsClkFreq",
            ctypes.c_uint32,
        ),  # clock frequency of cyclic message scheduler [Hz]
        ("dwCmsDivisor", ctypes.c_uint32),  # divisor for the cyclic message scheduler
        (
            "dwCmsMaxTicks",
            ctypes.c_uint32,
        ),  # maximum tick count value of the cyclic message
        (
            "dwDtxClkFreq",
            ctypes.c_uint32,
        ),  # clock frequency of the delayed message transmitter [Hz]
        (
            "dwDtxDivisor",
            ctypes.c_uint32,
        ),  # divisor for the delayed message transmitter
        (
            "dwDtxMaxTicks",
            ctypes.c_uint32,
        ),  # maximum tick count value of the delayed message transmitter
    ]


PCANCAPABILITIES2 = ctypes.POINTER(CANCAPABILITIES2)


class CANLINESTATUS2(ctypes.Structure):
    _fields_ = [
        ("bOpMode", ctypes.c_uint8),  # current CAN operating mode
        ("bExMode", ctypes.c_uint8),  # current CAN extended operating mode
        ("bBusLoad", ctypes.c_uint8),  # average bus load in percent (0..100)
        ("bReserved", ctypes.c_uint8),  # reserved set to 0
        ("sBtpSdr", ctypes.c_uint8),  # standard bit rate timing
        ("sBtpFdr", ctypes.c_uint8),  # fast data bit rate timing
        ("dwStatus", ctypes.c_uint32),  # status of the CAN controller (see CAN_STATUS_)
    ]


PCANLINESTATUS2 = ctypes.POINTER(CANLINESTATUS2)


class CANMSG2(ctypes.Structure):
    _fields_ = [
        ("dwTime", ctypes.c_uint32),  # time stamp for receive message
        ("rsvd", ctypes.c_uint32),  # reserved (set to 0)
        ("dwMsgId", ctypes.c_uint32),  # CAN message identifier (INTEL format)
        ("uMsgInfo", CANMSGINFO),  # message information (bit field)
        ("abData", ctypes.c_uint8 * 64),  # message data
    ]


PCANMSG2 = ctypes.POINTER(CANMSG2)


class CANCYCLICTXMSG2(ctypes.Structure):
    _fields_ = [
        ("wCycleTime", ctypes.c_uint16),  # cycle time for the message in ticks
        (
            "bIncrMode",
            ctypes.c_uint8,
        ),  # auto increment mode (see CAN_CTXMSG_INC_ const)
        (
            "bByteIndex",
            ctypes.c_uint8,
        ),  # index of the byte within abData[] to increment
        ("dwMsgId", ctypes.c_uint32),  # message identifier (INTEL format)
        ("uMsgInfo", CANMSGINFO),  # message information (bit field)
        ("abData", ctypes.c_uint8 * 64),  # message data
    ]


PCANCYCLICTXMSG2 = ctypes.POINTER(CANCYCLICTXMSG2)



#!/usr/bin/env python

import struct

import pytest

import can
from can.interfaces.pcan.pcan import PCAN_BITRATES


def test_sja1000():
    """Test some values obtained using other bit timing calculators."""
    timing = can.BitTiming(
        f_clock=8_000_000, brp=4, tseg1=11, tseg2=4, sjw=2, nof_samples=3
    )
    assert timing.f_clock == 8_000_000
    assert timing.bitrate == 125_000
    assert timing.brp == 4
    assert timing.nbt == 16
    assert timing.tseg1 == 11
    assert timing.tseg2 == 4
    assert timing.sjw == 2
    assert timing.nof_samples == 3
    assert timing.sample_point == 75
    assert timing.btr0 == 0x43
    assert timing.btr1 == 0xBA

    timing = can.BitTiming(f_clock=8_000_000, brp=1, tseg1=13, tseg2=2, sjw=1)
    assert timing.f_clock == 8_000_000
    assert timing.bitrate == 500_000
    assert timing.brp == 1
    assert timing.nbt == 16
    assert timing.tseg1 == 13
    assert timing.tseg2 == 2
    assert timing.sjw == 1
    assert timing.nof_samples == 1
    assert timing.sample_point == 87.5
    assert timing.btr0 == 0x00
    assert timing.btr1 == 0x1C

    timing = can.BitTiming(f_clock=8_000_000, brp=1, tseg1=5, tseg2=2, sjw=1)
    assert timing.f_clock == 8_000_000
    assert timing.bitrate == 1_000_000
    assert timing.brp == 1
    assert timing.nbt == 8
    assert timing.tseg1 == 5
    assert timing.tseg2 == 2
    assert timing.sjw == 1
    assert timing.nof_samples == 1
    assert timing.sample_point == 75
    assert timing.btr0 == 0x00
    assert timing.btr1 == 0x14


def test_from_bitrate_and_segments():
    timing = can.BitTiming.from_bitrate_and_segments(
        f_clock=8_000_000, bitrate=125_000, tseg1=11, tseg2=4, sjw=2, nof_samples=3
    )
    assert timing.f_clock == 8_000_000
    assert timing.bitrate == 125_000
    assert timing.brp == 4
    assert timing.nbt == 16
    assert timing.tseg1 == 11
    assert timing.tseg2 == 4
    assert timing.sjw == 2
    assert timing.nof_samples == 3
    assert timing.sample_point == 75
    assert timing.btr0 == 0x43
    assert timing.btr1 == 0xBA

    timing = can.BitTiming.from_bitrate_and_segments(
        f_clock=8_000_000, bitrate=500_000, tseg1=13, tseg2=2, sjw=1
    )
    assert timing.f_clock == 8_000_000
    assert timing.bitrate == 500_000
    assert timing.brp == 1
    assert timing.nbt == 16
    assert timing.tseg1 == 13
    assert timing.tseg2 == 2
    assert timing.sjw == 1
    assert timing.nof_samples == 1
    assert timing.sample_point == 87.5
    assert timing.btr0 == 0x00
    assert timing.btr1 == 0x1C

    timing = can.BitTiming.from_bitrate_and_segments(
        f_clock=8_000_000, bitrate=1_000_000, tseg1=5, tseg2=2, sjw=1
    )
    assert timing.f_clock == 8_000_000
    assert timing.bitrate == 1_000_000
    assert timing.brp == 1
    assert timing.nbt == 8
    assert timing.tseg1 == 5
    assert timing.tseg2 == 2
    assert timing.sjw == 1
    assert timing.nof_samples == 1
    assert timing.sample_point == 75
    assert timing.btr0 == 0x00
    assert timing.btr1 == 0x14

    timing = can.BitTimingFd.from_bitrate_and_segments(
        f_clock=80_000_000,
        nom_bitrate=500_000,
        nom_tseg1=119,
        nom_tseg2=40,
        nom_sjw=40,
        data_bitrate=2_000_000,
        data_tseg1=29,
        data_tseg2=10,
        data_sjw=10,
    )

    assert timing.f_clock == 80_000_000
    assert timing.nom_bitrate == 500_000
    assert timing.nom_brp == 1
    assert timing.nbt == 160
    assert timing.nom_tseg1 == 119
    assert timing.nom_tseg2 == 40
    assert timing.nom_sjw == 40
    assert timing.nom_sample_point == 75
    assert timing.f_clock == 80_000_000
    assert timing.data_bitrate == 2_000_000
    assert timing.data_brp == 1
    assert timing.dbt == 40
    assert timing.data_tseg1 == 29
    assert timing.data_tseg2 == 10
    assert timing.data_sjw == 10
    assert timing.data_sample_point == 75


def test_can_fd():
    timing = can.BitTimingFd(
        f_clock=80_000_000,
        nom_brp=1,
        nom_tseg1=119,
        nom_tseg2=40,
        nom_sjw=40,
        data_brp=1,
        data_tseg1=29,
        data_tseg2=10,
        data_sjw=10,
    )

    assert timing.f_clock == 80_000_000
    assert timing.nom_bitrate == 500_000
    assert timing.nom_brp == 1
    assert timing.nbt == 160
    assert timing.nom_tseg1 == 119
    assert timing.nom_tseg2 == 40
    assert timing.nom_sjw == 40
    assert timing.nom_sample_point == 75
    assert timing.f_clock == 80_000_000
    assert timing.data_bitrate == 2_000_000
    assert timing.data_brp == 1
    assert timing.dbt == 40
    assert timing.data_tseg1 == 29
    assert timing.data_tseg2 == 10
    assert timing.data_sjw == 10
    assert timing.data_sample_point == 75


def test_from_btr():
    timing = can.BitTiming.from_registers(f_clock=8_000_000, btr0=0x00, btr1=0x14)
    assert timing.bitrate == 1_000_000
    assert timing.brp == 1
    assert timing.nbt == 8
    assert timing.tseg1 == 5
    assert timing.tseg2 == 2
    assert timing.sjw == 1
    assert timing.sample_point == 75
    assert timing.btr0 == 0x00
    assert timing.btr1 == 0x14


def test_btr_persistence():
    f_clock = 8_000_000
    for btr0btr1 in PCAN_BITRATES.values():
        btr1, btr0 = struct.unpack("BB", btr0btr1)

        t = can.BitTiming.from_registers(f_clock, btr0, btr1)
        assert t.btr0 == btr0
        assert t.btr1 == btr1


def test_from_sample_point():
    timing = can.BitTiming.from_sample_point(
        f_clock=16_000_000,
        bitrate=500_000,
        sample_point=69.0,
    )
    assert timing.f_clock == 16_000_000
    assert timing.bitrate == 500_000
    assert 68 < timing.sample_point < 70

    fd_timing = can.BitTimingFd.from_sample_point(
        f_clock=80_000_000,
        nom_bitrate=1_000_000,
        nom_sample_point=75.0,
        data_bitrate=8_000_000,
        data_sample_point=70.0,
    )
    assert fd_timing.f_clock == 80_000_000
    assert fd_timing.nom_bitrate == 1_000_000
    assert 74 < fd_timing.nom_sample_point < 76
    assert fd_timing.data_bitrate == 8_000_000
    assert 69 < fd_timing.data_sample_point < 71

    # check that there is a solution for every sample point
    for sp in range(50, 100):
        can.BitTiming.from_sample_point(
            f_clock=16_000_000, bitrate=500_000, sample_point=sp
        )

    # check that there is a solution for every sample point
    for nsp in range(50, 100):
        for dsp in range(50, 100):
            can.BitTimingFd.from_sample_point(
                f_clock=80_000_000,
                nom_bitrate=500_000,
                nom_sample_point=nsp,
                data_bitrate=2_000_000,
                data_sample_point=dsp,
            )


def test_equality():
    t1 = can.BitTiming.from_registers(f_clock=8_000_000, btr0=0x00, btr1=0x14)
    t2 = can.BitTiming(f_clock=8_000_000, brp=1, tseg1=5, tseg2=2, sjw=1, nof_samples=1)
    t3 = can.BitTiming(
        f_clock=16_000_000, brp=2, tseg1=5, tseg2=2, sjw=1, nof_samples=1
    )
    assert t1 == t2
    assert t1 != t3
    assert t2 != t3
    assert t1 != 10

    t4 = can.BitTimingFd(
        f_clock=80_000_000,
        nom_brp=1,
        nom_tseg1=119,
        nom_tseg2=40,
        nom_sjw=40,
        data_brp=1,
        data_tseg1=29,
        data_tseg2=10,
        data_sjw=10,
    )
    t5 = can.BitTimingFd(
        f_clock=80_000_000,
        nom_brp=1,
        nom_tseg1=119,
        nom_tseg2=40,
        nom_sjw=40,
        data_brp=1,
        data_tseg1=29,
        data_tseg2=10,
        data_sjw=10,
    )
    t6 = can.BitTimingFd.from_sample_point(
        f_clock=80_000_000,
        nom_bitrate=1_000_000,
        nom_sample_point=75.0,
        data_bitrate=8_000_000,
        data_sample_point=70.0,
    )
    assert t4 == t5
    assert t4 != t6
    assert t4 != t1


def test_string_representation():
    timing = can.BitTiming(f_clock=8_000_000, brp=1, tseg1=5, tseg2=2, sjw=1)
    assert str(timing) == (
        "BR: 1_000_000 bit/s, SP: 75.00%, BRP: 1, TSEG1: 5, TSEG2: 2, SJW: 1, "
        "BTR: 0014h, CLK: 8MHz"
    )

    fd_timing = can.BitTimingFd(
        f_clock=80_000_000,
        nom_brp=1,
        nom_tseg1=119,
        nom_tseg2=40,
        nom_sjw=40,
        data_brp=1,
        data_tseg1=29,
        data_tseg2=10,
        data_sjw=10,
    )
    assert str(fd_timing) == (
        "NBR: 500_000 bit/s, NSP: 75.00%, NBRP: 1, NTSEG1: 119, NTSEG2: 40, NSJW: 40, "
        "DBR: 2_000_000 bit/s, DSP: 75.00%, DBRP: 1, DTSEG1: 29, DTSEG2: 10, DSJW: 10, "
        "CLK: 80MHz"
    )


def test_repr():
    timing = can.BitTiming(f_clock=8_000_000, brp=1, tseg1=5, tseg2=2, sjw=1)
    assert repr(timing) == (
        "can.BitTiming(f_clock=8000000, brp=1, tseg1=5, tseg2=2, sjw=1, nof_samples=1)"
    )

    fd_timing = can.BitTimingFd(
        f_clock=80_000_000,
        nom_brp=1,
        nom_tseg1=119,
        nom_tseg2=40,
        nom_sjw=40,
        data_brp=1,
        data_tseg1=29,
        data_tseg2=10,
        data_sjw=10,
    )
    assert repr(fd_timing) == (
        "can.BitTimingFd(f_clock=80000000, nom_brp=1, nom_tseg1=119, nom_tseg2=40, "
        "nom_sjw=40, data_brp=1, data_tseg1=29, data_tseg2=10, data_sjw=10)"
    )


def test_hash():
    _timings = {
        can.BitTiming(f_clock=8_000_000, brp=1, tseg1=5, tseg2=2, sjw=1, nof_samples=1),
        can.BitTimingFd(
            f_clock=80_000_000,
            nom_brp=1,
            nom_tseg1=119,
            nom_tseg2=40,
            nom_sjw=40,
            data_brp=1,
            data_tseg1=29,
            data_tseg2=10,
            data_sjw=10,
        ),
    }


def test_mapping():
    timing = can.BitTiming(f_clock=8_000_000, brp=1, tseg1=5, tseg2=2, sjw=1)
    timing_dict = dict(timing)
    assert timing_dict["f_clock"] == timing["f_clock"]
    assert timing_dict["brp"] == timing["brp"]
    assert timing_dict["tseg1"] == timing["tseg1"]
    assert timing_dict["tseg2"] == timing["tseg2"]
    assert timing_dict["sjw"] == timing["sjw"]
    assert timing == can.BitTiming(**timing_dict)

    fd_timing = can.BitTimingFd(
        f_clock=80_000_000,
        nom_brp=1,
        nom_tseg1=119,
        nom_tseg2=40,
        nom_sjw=40,
        data_brp=1,
        data_tseg1=29,
        data_tseg2=10,
        data_sjw=10,
    )
    fd_timing_dict = dict(fd_timing)
    assert fd_timing_dict["f_clock"] == fd_timing["f_clock"]
    assert fd_timing_dict["nom_brp"] == fd_timing["nom_brp"]
    assert fd_timing_dict["nom_tseg1"] == fd_timing["nom_tseg1"]
    assert fd_timing_dict["nom_tseg2"] == fd_timing["nom_tseg2"]
    assert fd_timing_dict["nom_sjw"] == fd_timing["nom_sjw"]
    assert fd_timing_dict["data_brp"] == fd_timing["data_brp"]
    assert fd_timing_dict["data_tseg1"] == fd_timing["data_tseg1"]
    assert fd_timing_dict["data_tseg2"] == fd_timing["data_tseg2"]
    assert fd_timing_dict["data_sjw"] == fd_timing["data_sjw"]
    assert fd_timing == can.BitTimingFd(**fd_timing_dict)


def test_oscillator_tolerance():
    timing = can.BitTiming(f_clock=16_000_000, brp=2, tseg1=10, tseg2=5, sjw=4)
    osc_tol = timing.oscillator_tolerance(
        node_loop_delay_ns=250,
        bus_length_m=10.0,
    )
    assert osc_tol == pytest.approx(1.23, abs=1e-2)

    fd_timing = can.BitTimingFd(
        f_clock=80_000_000,
        nom_brp=5,
        nom_tseg1=27,
        nom_tseg2=4,
        nom_sjw=4,
        data_brp=5,
        data_tseg1=6,
        data_tseg2=1,
        data_sjw=1,
    )
    osc_tol = fd_timing.oscillator_tolerance(
        node_loop_delay_ns=250,
        bus_length_m=10.0,
    )
    assert osc_tol == pytest.approx(0.48, abs=1e-2)


def test_recreate_with_f_clock():
    timing_8mhz = can.BitTiming(f_clock=8_000_000, brp=1, tseg1=5, tseg2=2, sjw=1)
    timing_16mhz = timing_8mhz.recreate_with_f_clock(f_clock=16_000_000)
    assert timing_8mhz.bitrate == timing_16mhz.bitrate
    assert timing_8mhz.sample_point == timing_16mhz.sample_point
    assert (timing_8mhz.sjw / timing_8mhz.nbt) == pytest.approx(
        timing_16mhz.sjw / timing_16mhz.nbt, abs=1e-3
    )
    assert timing_8mhz.nof_samples == timing_16mhz.nof_samples

    timing_16mhz = can.BitTiming(
        f_clock=16000000, brp=2, tseg1=12, tseg2=3, sjw=3, nof_samples=1
    )
    timing_8mhz = timing_16mhz.recreate_with_f_clock(f_clock=8_000_000)
    assert timing_8mhz.bitrate == timing_16mhz.bitrate
    assert timing_8mhz.sample_point == timing_16mhz.sample_point
    assert (timing_8mhz.sjw / timing_8mhz.nbt) == pytest.approx(
        timing_16mhz.sjw / timing_16mhz.nbt, abs=1e-2
    )
    assert timing_8mhz.nof_samples == timing_16mhz.nof_samples

    fd_timing_80mhz = can.BitTimingFd(
        f_clock=80_000_000,
        nom_brp=5,
        nom_tseg1=27,
        nom_tseg2=4,
        nom_sjw=4,
        data_brp=5,
        data_tseg1=6,
        data_tseg2=1,
        data_sjw=1,
    )
    fd_timing_60mhz = fd_timing_80mhz.recreate_with_f_clock(f_clock=60_000_000)
    assert fd_timing_80mhz.nom_bitrate == fd_timing_60mhz.nom_bitrate
    assert fd_timing_80mhz.nom_sample_point == pytest.approx(
        fd_timing_60mhz.nom_sample_point, abs=1.0
    )
    assert (fd_timing_80mhz.nom_sjw / fd_timing_80mhz.nbt) == pytest.approx(
        fd_timing_60mhz.nom_sjw / fd_timing_60mhz.nbt, abs=1e-2
    )
    assert fd_timing_80mhz.data_bitrate == fd_timing_60mhz.data_bitrate
    assert fd_timing_80mhz.data_sample_point == pytest.approx(
        fd_timing_60mhz.data_sample_point, abs=1.0
    )
    assert (fd_timing_80mhz.data_sjw / fd_timing_80mhz.dbt) == pytest.approx(
        fd_timing_60mhz.data_sjw / fd_timing_60mhz.dbt, abs=1e-2
    )



from ctypes import c_ubyte as BYTE
from ctypes import c_ulong as DWORD
from ctypes import c_ushort as WORD

#: Maximum number of modules that are supported.
MAX_MODULES = 64

#: Maximum number of applications that can use the USB-CAN-library.
MAX_INSTANCES = 64

#: With the method :meth:`UcanServer.init_can` the module is used, which is detected at first.
#: This value only should be used in case only one module is connected to the computer.
ANY_MODULE = 255

#: No valid USB-CAN Handle (only used internally).
INVALID_HANDLE = 0xFF


class Baudrate(WORD):
    """
    Specifies pre-defined baud rate values for GW-001, GW-002 and all systec USB-CANmoduls.

    .. seealso::

       :meth:`UcanServer.init_can`

       :meth:`UcanServer.set_baudrate`

       :meth:`UcanServer.get_baudrate_message`

       :class:`BaudrateEx`
    """

    #: 1000 kBit/sec
    BAUD_1MBit = 0x14
    #: 800 kBit/sec
    BAUD_800kBit = 0x16
    #: 500 kBit/sec
    BAUD_500kBit = 0x1C
    #: 250 kBit/sec
    BAUD_250kBit = 0x11C
    #: 125 kBit/sec
    BAUD_125kBit = 0x31C
    #: 100 kBit/sec
    BAUD_100kBit = 0x432F
    #: 50 kBit/sec
    BAUD_50kBit = 0x472F
    #: 20 kBit/sec
    BAUD_20kBit = 0x532F
    #: 10 kBit/sec
    BAUD_10kBit = 0x672F
    #: Uses pre-defined extended values of baudrate for all systec USB-CANmoduls.
    BAUD_USE_BTREX = 0x0
    #: Automatic baud rate detection (not implemented in this version).
    BAUD_AUTO = -1


class BaudrateEx(DWORD):
    """
    Specifies pre-defined baud rate values for all systec USB-CANmoduls.

    These values cannot be used for GW-001 and GW-002! Use values from enum :class:`Baudrate` instead.

    .. seealso::

       :meth:`UcanServer.init_can`

       :meth:`UcanServer.set_baudrate`

       :meth:`UcanServer.get_baudrate_ex_message`

       :class:`Baudrate`
    """

    #: G3: 1000 kBit/sec
    BAUDEX_1MBit = 0x20354
    #: G3: 800 kBit/sec
    BAUDEX_800kBit = 0x30254
    #: G3: 500 kBit/sec
    BAUDEX_500kBit = 0x50354
    #: G3: 250 kBit/sec
    BAUDEX_250kBit = 0xB0354
    #: G3: 125 kBit/sec
    BAUDEX_125kBit = 0x170354
    #: G3: 100 kBit/sec
    BAUDEX_100kBit = 0x170466
    #: G3: 50 kBit/sec
    BAUDEX_50kBit = 0x2F0466
    #: G3: 20 kBit/sec
    BAUDEX_20kBit = 0x770466
    #: G3: 10 kBit/sec (half CPU clock)
    BAUDEX_10kBit = 0x80770466
    #: G3: 1000 kBit/sec Sample Point: 87,50%
    BAUDEX_SP2_1MBit = 0x20741
    #: G3: 800 kBit/sec Sample Point: 86,67%
    BAUDEX_SP2_800kBit = 0x30731
    #: G3: 500 kBit/sec Sample Point: 87,50%
    BAUDEX_SP2_500kBit = 0x50741
    #: G3: 250 kBit/sec Sample Point: 87,50%
    BAUDEX_SP2_250kBit = 0xB0741
    #: G3: 125 kBit/sec Sample Point: 87,50%
    BAUDEX_SP2_125kBit = 0x170741
    #: G3: 100 kBit/sec Sample Point: 87,50%
    BAUDEX_SP2_100kBit = 0x1D1741
    #: G3: 50 kBit/sec Sample Point: 87,50%
    BAUDEX_SP2_50kBit = 0x3B1741
    #: G3: 20 kBit/sec Sample Point: 85,00%
    BAUDEX_SP2_20kBit = 0x771772
    #: G3: 10 kBit/sec Sample Point: 85,00% (half CPU clock)
    BAUDEX_SP2_10kBit = 0x80771772

    #: G4: 1000 kBit/sec Sample Point: 83,33%
    BAUDEX_G4_1MBit = 0x406F0000
    #: G4: 800 kBit/sec Sample Point: 80,00%
    BAUDEX_G4_800kBit = 0x402A0001
    #: G4: 500 kBit/sec Sample Point: 83,33%
    BAUDEX_G4_500kBit = 0x406F0001
    #: G4: 250 kBit/sec Sample Point: 83,33%
    BAUDEX_G4_250kBit = 0x406F0003
    #: G4: 125 kBit/sec Sample Point: 83,33%
    BAUDEX_G4_125kBit = 0x406F0007
    #: G4: 100 kBit/sec Sample Point: 83,33%
    BAUDEX_G4_100kBit = 0x416F0009
    #: G4: 50 kBit/sec Sample Point: 83,33%
    BAUDEX_G4_50kBit = 0x416F0013
    #: G4: 20 kBit/sec Sample Point: 84,00%
    BAUDEX_G4_20kBit = 0x417F002F
    #: G4: 10 kBit/sec Sample Point: 84,00% (half CPU clock)
    BAUDEX_G4_10kBit = 0x417F005F
    #: Uses pre-defined values of baud rates of :class:`Baudrate`.
    BAUDEX_USE_BTR01 = 0x0
    #: Automatic baud rate detection (not implemented in this version).
    BAUDEX_AUTO = 0xFFFFFFFF


class MsgFrameFormat(BYTE):
    """
    Specifies values for the frame format of CAN messages for member :attr:`CanMsg.m_bFF` in structure
    :class:`CanMsg`. These values can be combined.

    .. seealso:: :class:`CanMsg`
    """

    #: standard CAN data frame with 11 bit ID (CAN2.0A spec.)
    MSG_FF_STD = 0x0
    #: transmit echo
    MSG_FF_ECHO = 0x20
    #: CAN remote request frame with
    MSG_FF_RTR = 0x40
    #: extended CAN data frame with 29 bit ID (CAN2.0B spec.)
    MSG_FF_EXT = 0x80


class ReturnCode(BYTE):
    """
    Specifies all return codes of all methods of this class.
    """

    #: no error
    SUCCESSFUL = 0x0
    # start of error codes coming from USB-CAN-library
    ERR = 0x1
    # start of error codes coming from command interface between host and USB-CANmodul
    ERRCMD = 0x40
    # start of warning codes
    WARNING = 0x80
    # start of reserved codes which are only used internally
    RESERVED = 0xC0

    #: could not created a resource (memory, handle, ...)
    ERR_RESOURCE = 0x1
    #: the maximum number of opened modules is reached
    ERR_MAXMODULES = 0x2
    #: the specified module is already in use
    ERR_HWINUSE = 0x3
    #: the software versions of the module and library are incompatible
    ERR_ILLVERSION = 0x4
    #: the module with the specified device number is not connected (or used by an other application)
    ERR_ILLHW = 0x5
    #: wrong USB-CAN-Handle handed over to the function
    ERR_ILLHANDLE = 0x6
    #: wrong parameter handed over to the function
    ERR_ILLPARAM = 0x7
    #: instruction can not be processed at this time
    ERR_BUSY = 0x8
    #: no answer from module
    ERR_TIMEOUT = 0x9
    #: a request to the driver failed
    ERR_IOFAILED = 0xA
    #: a CAN message did not fit into the transmit buffer
    ERR_DLL_TXFULL = 0xB
    #: maximum number of applications is reached
    ERR_MAXINSTANCES = 0xC
    #: CAN interface is not yet initialized
    ERR_CANNOTINIT = 0xD
    #: USB-CANmodul was disconnected
    ERR_DISCONECT = 0xE
    #: the needed device class does not exist
    ERR_NOHWCLASS = 0xF
    #: illegal CAN channel
    ERR_ILLCHANNEL = 0x10
    #: reserved
    ERR_RESERVED1 = 0x11
    #: the API function can not be used with this hardware
    ERR_ILLHWTYPE = 0x12

    #: the received response does not match to the transmitted command
    ERRCMD_NOTEQU = 0x40
    #: no access to the CAN controller
    ERRCMD_REGTST = 0x41
    #: the module could not interpret the command
    ERRCMD_ILLCMD = 0x42
    #: error while reading the EEPROM
    ERRCMD_EEPROM = 0x43
    #: reserved
    ERRCMD_RESERVED1 = 0x44
    #: reserved
    ERRCMD_RESERVED2 = 0x45
    #: reserved
    ERRCMD_RESERVED3 = 0x46
    #: illegal baud rate value specified in BTR0/BTR1 for systec USB-CANmoduls
    ERRCMD_ILLBDR = 0x47
    #: CAN channel is not initialized
    ERRCMD_NOTINIT = 0x48
    #: CAN channel is already initialized
    ERRCMD_ALREADYINIT = 0x49
    #: illegal sub-command specified
    ERRCMD_ILLSUBCMD = 0x4A
    #: illegal index specified (e.g. index for cyclic CAN messages)
    ERRCMD_ILLIDX = 0x4B
    #: cyclic CAN message(s) can not be defined because transmission of cyclic CAN messages is already running
    ERRCMD_RUNNING = 0x4C

    #: no CAN messages received
    WARN_NODATA = 0x80
    #: overrun in receive buffer of the kernel driver
    WARN_SYS_RXOVERRUN = 0x81
    #: overrun in receive buffer of the USB-CAN-library
    WARN_DLL_RXOVERRUN = 0x82
    #: reserved
    WARN_RESERVED1 = 0x83
    #: reserved
    WARN_RESERVED2 = 0x84
    #: overrun in transmit buffer of the firmware (but this CAN message was successfully stored in buffer of the
    #: library)
    WARN_FW_TXOVERRUN = 0x85
    #: overrun in receive buffer of the firmware (but this CAN message was successfully read)
    WARN_FW_RXOVERRUN = 0x86
    #: reserved
    WARN_FW_TXMSGLOST = 0x87
    #: pointer is NULL
    WARN_NULL_PTR = 0x90
    #: not all CAN messages could be stored to the transmit buffer in USB-CAN-library (check output of parameter
    #: pdwCount_p)
    WARN_TXLIMIT = 0x91
    #: reserved
    WARN_BUSY = 0x92


class CbEvent(BYTE):
    """
    This enum defines events for the callback functions of the library.

    .. seealso:: :meth:`UcanServer.get_status`
    """

    #: The USB-CANmodul has been initialized.
    EVENT_INITHW = 0
    #: The CAN interface has been initialized.
    EVENT_init_can = 1
    #: A new CAN message has been received.
    EVENT_RECEIVE = 2
    #: The error state in the module has changed.
    EVENT_STATUS = 3
    #: The CAN interface has been deinitialized.
    EVENT_DEINIT_CAN = 4
    #: The USB-CANmodul has been deinitialized.
    EVENT_DEINITHW = 5
    #: A new USB-CANmodul has been connected.
    EVENT_CONNECT = 6
    #: Any USB-CANmodul has been disconnected.
    EVENT_DISCONNECT = 7
    #: A USB-CANmodul has been disconnected during operation.
    EVENT_FATALDISCON = 8
    #: Reserved
    EVENT_RESERVED1 = 0x80


class CanStatus(WORD):
    """
    CAN error status bits. These bit values occurs in combination with the method :meth:`UcanServer.get_status`.

    .. seealso::

       :meth:`UcanServer.get_status`

       :meth:`UcanServer.get_can_status_message`
    """

    #: No error.
    CANERR_OK = 0x0
    #: Transmit buffer of the CAN controller is full.
    CANERR_XMTFULL = 0x1
    #: Receive buffer of the CAN controller is full.
    CANERR_OVERRUN = 0x2
    #: Bus error: Error Limit 1 exceeded (Warning Limit reached)
    CANERR_BUSLIGHT = 0x4
    #: Bus error: Error Limit 2 exceeded (Error Passive)
    CANERR_BUSHEAVY = 0x8
    #: Bus error: CAN controller has gone into Bus-Off state.
    #: Method :meth:`UcanServer.reset_can` has to be called.
    CANERR_BUSOFF = 0x10
    #: No CAN message is within the receive buffer.
    CANERR_QRCVEMPTY = 0x20
    #: Receive buffer is full. CAN messages has been lost.
    CANERR_QOVERRUN = 0x40
    #: Transmit buffer is full.
    CANERR_QXMTFULL = 0x80
    #: Register test of the CAN controller failed.
    CANERR_REGTEST = 0x100
    #: Memory test on hardware failed.
    CANERR_MEMTEST = 0x200
    #: Transmit CAN message(s) was/were automatically deleted by firmware (transmit timeout).
    CANERR_TXMSGLOST = 0x400


class UsbStatus(WORD):
    """
    USB error status bits. These bit values occurs in combination with the method :meth:`UcanServer.get_status`.

    .. seealso:: :meth:`UcanServer.get_status`
    """

    #: No error.
    USBERR_OK = 0x0


#: Specifies the acceptance mask for receiving all CAN messages.
#:
#: .. seealso::
#:
#:    :const:`ACR_ALL`
#:
#:    :meth:`UcanServer.init_can`
#:
#:    :meth:`UcanServer.set_acceptance`
AMR_ALL = 0xFFFFFFFF

#: Specifies the acceptance code for receiving all CAN messages.
#:
#: .. seealso::
#:
#:    :const:`AMR_ALL`
#:
#:    :meth:`UcanServer.init_can`
#:
#:    :meth:`UcanServer.set_acceptance`
ACR_ALL = 0x0


class OutputControl(BYTE):
    """
    Specifies pre-defined values for the Output Control Register of SJA1000 on GW-001 and GW-002.
    These values are only important for GW-001 and GW-002.
    They does not have an effect on systec USB-CANmoduls.
    """

    #: default OCR value for the standard USB-CANmodul GW-001/GW-002
    OCR_DEFAULT = 0x1A
    #: OCR value for RS485 interface and galvanic isolation
    OCR_RS485_ISOLATED = 0x1E
    #: OCR value for RS485 interface but without galvanic isolation
    OCR_RS485_NOT_ISOLATED = 0xA


#: Specifies the default value for the maximum number of entries in the receive and transmit buffer.
DEFAULT_BUFFER_ENTRIES = 4096


class Channel(BYTE):
    """
    Specifies values for the CAN channel to be used on multi-channel USB-CANmoduls.
    """

    #: Specifies the first CAN channel (GW-001/GW-002 and USB-CANmodul1 only can be used with this channel).
    CHANNEL_CH0 = 0
    #: Specifies the second CAN channel (this channel cannot be used with GW-001/GW-002 and USB-CANmodul1).
    CHANNEL_CH1 = 1
    #: Specifies all CAN channels (can only be used with the method :meth:`UcanServer.shutdown`).
    CHANNEL_ALL = 254
    #: Specifies the use of any channel (can only be used with the method :meth:`UcanServer.read_can_msg`).
    CHANNEL_ANY = 255
    #: Specifies the first CAN channel (equivalent to :data:`CHANNEL_CH0`).
    CHANNEL_CAN1 = CHANNEL_CH0
    #: Specifies the second CAN channel (equivalent to :data:`CHANNEL_CH1`).
    CHANNEL_CAN2 = CHANNEL_CH1
    #: Specifies the LIN channel (currently not supported by the software).
    CHANNEL_LIN = CHANNEL_CH1


class ResetFlags(DWORD):
    """
    Specifies flags for resetting USB-CANmodul with method :meth:`UcanServer.reset_can`.
    These flags can be used in combination.

    .. seealso:: :meth:`UcanServer.reset_can`
    """

    #: reset everything
    RESET_ALL = 0x0
    #: no CAN status reset (only supported for systec USB-CANmoduls)
    RESET_NO_STATUS = 0x1
    #: no CAN controller reset
    RESET_NO_CANCTRL = 0x2
    #: no transmit message counter reset
    RESET_NO_TXCOUNTER = 0x4
    #: no receive message counter reset
    RESET_NO_RXCOUNTER = 0x8
    #: no transmit message buffer reset at channel level
    RESET_NO_TXBUFFER_CH = 0x10
    #: no transmit message buffer reset at USB-CAN-library level
    RESET_NO_TXBUFFER_DLL = 0x20
    #: no transmit message buffer reset at firmware level
    RESET_NO_TXBUFFER_FW = 0x80
    #: no receive message buffer reset at channel level
    RESET_NO_RXBUFFER_CH = 0x100
    #: no receive message buffer reset at USB-CAN-library level
    RESET_NO_RXBUFFER_DLL = 0x200
    #: no receive message buffer reset at kernel driver level
    RESET_NO_RXBUFFER_SYS = 0x400
    #: no receive message buffer reset at firmware level
    RESET_NO_RXBUFFER_FW = 0x800
    #: complete firmware reset (module will automatically reconnect at USB port in 500msec)
    RESET_FIRMWARE = 0xFFFFFFFF

    #: no reset of all message counters
    RESET_NO_COUNTER_ALL = RESET_NO_TXCOUNTER | RESET_NO_RXCOUNTER
    #: no reset of transmit message buffers at communication level (firmware, kernel and library)
    RESET_NO_TXBUFFER_COMM = RESET_NO_TXBUFFER_DLL | 0x40 | RESET_NO_TXBUFFER_FW
    #: no reset of receive message buffers at communication level (firmware, kernel and library)
    RESET_NO_RXBUFFER_COMM = (
        RESET_NO_RXBUFFER_DLL | RESET_NO_RXBUFFER_SYS | RESET_NO_RXBUFFER_FW
    )
    #: no reset of all transmit message buffers
    RESET_NO_TXBUFFER_ALL = RESET_NO_TXBUFFER_CH | RESET_NO_TXBUFFER_COMM
    #: no reset of all receive message buffers
    RESET_NO_RXBUFFER_ALL = RESET_NO_RXBUFFER_CH | RESET_NO_RXBUFFER_COMM
    #: no reset of all message buffers at communication level (firmware, kernel and library)
    RESET_NO_BUFFER_COMM = RESET_NO_TXBUFFER_COMM | RESET_NO_RXBUFFER_COMM
    #: no reset of all message buffers
    RESET_NO_BUFFER_ALL = RESET_NO_TXBUFFER_ALL | RESET_NO_RXBUFFER_ALL
    #: reset of the CAN status only
    RESET_ONLY_STATUS = 0xFFFF & ~RESET_NO_STATUS
    #: reset of the CAN controller only
    RESET_ONLY_CANCTRL = 0xFFFF & ~RESET_NO_CANCTRL
    #: reset of the transmit buffer in firmware only
    RESET_ONLY_TXBUFFER_FW = 0xFFFF & ~RESET_NO_TXBUFFER_FW
    #: reset of the receive buffer in firmware only
    RESET_ONLY_RXBUFFER_FW = 0xFFFF & ~RESET_NO_RXBUFFER_FW
    #: reset of the specified channel of the receive buffer only
    RESET_ONLY_RXCHANNEL_BUFF = 0xFFFF & ~RESET_NO_RXBUFFER_CH
    #: reset of the specified channel of the transmit buffer only
    RESET_ONLY_TXCHANNEL_BUFF = 0xFFFF & ~RESET_NO_TXBUFFER_CH
    #: reset of the receive buffer and receive message counter only
    RESET_ONLY_RX_BUFF = 0xFFFF & ~(RESET_NO_RXBUFFER_ALL | RESET_NO_RXCOUNTER)
    #: reset of the receive buffer and receive message counter (for GW-002) only
    RESET_ONLY_RX_BUFF_GW002 = 0xFFFF & ~(
        RESET_NO_RXBUFFER_ALL | RESET_NO_RXCOUNTER | RESET_NO_TXBUFFER_FW
    )
    #: reset of the transmit buffer and transmit message counter only
    RESET_ONLY_TX_BUFF = 0xFFFF & ~(RESET_NO_TXBUFFER_ALL | RESET_NO_TXCOUNTER)
    #: reset of all buffers and all message counters only
    RESET_ONLY_ALL_BUFF = RESET_ONLY_RX_BUFF & RESET_ONLY_TX_BUFF
    #: reset of all message counters only
    RESET_ONLY_ALL_COUNTER = 0xFFFF & ~RESET_NO_COUNTER_ALL


PRODCODE_PID_TWO_CHA = 0x1
PRODCODE_PID_TERM = 0x1
PRODCODE_PID_RBUSER = 0x1
PRODCODE_PID_RBCAN = 0x1
PRODCODE_PID_G4 = 0x20
PRODCODE_PID_RESVD = 0x40

PRODCODE_MASK_DID = 0xFFFF0000
PRODCODE_MASK_PID = 0xFFFF
PRODCODE_MASK_PIDG3 = PRODCODE_MASK_PID & 0xFFFFFFBF


class ProductCode(WORD):
    """
    These values defines product codes for all known USB-CANmodul derivatives received in member
    :attr:`HardwareInfoEx.m_dwProductCode` of structure :class:`HardwareInfoEx`
    with method :meth:`UcanServer.get_hardware_info`.

    .. seealso::

       :meth:`UcanServer.get_hardware_info`

       :class:`HardwareInfoEx`
    """

    #: Product code for GW-001 (outdated).
    PRODCODE_PID_GW001 = 0x1100
    #: Product code for GW-002 (outdated).
    PRODCODE_PID_GW002 = 0x1102
    #: Product code for Multiport CAN-to-USB G3.
    PRODCODE_PID_MULTIPORT = 0x1103
    #: Product code for USB-CANmodul1 G3.
    PRODCODE_PID_BASIC = 0x1104
    #: Product code for USB-CANmodul2 G3.
    PRODCODE_PID_ADVANCED = 0x1105
    #: Product code for USB-CANmodul8 G3.
    PRODCODE_PID_USBCAN8 = 0x1107
    #: Product code for USB-CANmodul16 G3.
    PRODCODE_PID_USBCAN16 = 0x1109
    #: Reserved.
    PRODCODE_PID_RESERVED3 = 0x1110
    #: Product code for USB-CANmodul2 G4.
    PRODCODE_PID_ADVANCED_G4 = 0x1121
    #: Product code for USB-CANmodul1 G4.
    PRODCODE_PID_BASIC_G4 = 0x1122
    #: Reserved.
    PRODCODE_PID_RESERVED1 = 0x1144
    #: Reserved.
    PRODCODE_PID_RESERVED2 = 0x1145


#: Definitions for cyclic CAN messages.
MAX_CYCLIC_CAN_MSG = 16


class CyclicFlags(DWORD):
    """
    Specifies flags for cyclical CAN messages.
    These flags can be used in combinations with method :meth:`UcanServer.enable_cyclic_can_msg`.

    .. seealso:: :meth:`UcanServer.enable_cyclic_can_msg`
    """

    #: Stops the transmission of cyclic CAN messages.
    CYCLIC_FLAG_STOPP = 0x0
    #: Global enable of transmission of cyclic CAN messages.
    CYCLIC_FLAG_START = 0x80000000
    #: List of cyclic CAN messages will be processed in sequential mode (otherwise in parallel mode).
    CYCLIC_FLAG_SEQUMODE = 0x40000000
    #: No echo will be sent back if echo mode is enabled with method :meth:`UcanServer.init_can`.
    CYCLIC_FLAG_NOECHO = 0x10000
    #: CAN message with index 0 of the list will not be sent.
    CYCLIC_FLAG_LOCK_0 = 0x1
    #: CAN message with index 1 of the list will not be sent.
    CYCLIC_FLAG_LOCK_1 = 0x2
    #: CAN message with index 2 of the list will not be sent.
    CYCLIC_FLAG_LOCK_2 = 0x4
    #: CAN message with index 3 of the list will not be sent.
    CYCLIC_FLAG_LOCK_3 = 0x8
    #: CAN message with index 4 of the list will not be sent.
    CYCLIC_FLAG_LOCK_4 = 0x10
    #: CAN message with index 5 of the list will not be sent.
    CYCLIC_FLAG_LOCK_5 = 0x20
    #: CAN message with index 6 of the list will not be sent.
    CYCLIC_FLAG_LOCK_6 = 0x40
    #: CAN message with index 7 of the list will not be sent.
    CYCLIC_FLAG_LOCK_7 = 0x80
    #: CAN message with index 8 of the list will not be sent.
    CYCLIC_FLAG_LOCK_8 = 0x100
    #: CAN message with index 9 of the list will not be sent.
    CYCLIC_FLAG_LOCK_9 = 0x200
    #: CAN message with index 10 of the list will not be sent.
    CYCLIC_FLAG_LOCK_10 = 0x400
    #: CAN message with index 11 of the list will not be sent.
    CYCLIC_FLAG_LOCK_11 = 0x800
    #: CAN message with index 12 of the list will not be sent.
    CYCLIC_FLAG_LOCK_12 = 0x1000
    #: CAN message with index 13 of the list will not be sent.
    CYCLIC_FLAG_LOCK_13 = 0x2000
    #: CAN message with index 14 of the list will not be sent.
    CYCLIC_FLAG_LOCK_14 = 0x4000
    #: CAN message with index 15 of the list will not be sent.
    CYCLIC_FLAG_LOCK_15 = 0x8000


class PendingFlags(BYTE):
    """
    Specifies flags for method :meth:`UcanServer.get_msg_pending`.
    These flags can be uses in combinations.

    .. seealso:: :meth:`UcanServer.get_msg_pending`
    """

    #: number of pending CAN messages in receive buffer of USB-CAN-library
    PENDING_FLAG_RX_DLL = 0x1
    #: reserved
    PENDING_FLAG_RX_SYS = 0x2
    #: number of pending CAN messages in receive buffer of firmware
    PENDING_FLAG_RX_FW = 0x4
    #: number of pending CAN messages in transmit buffer of USB-CAN-library
    PENDING_FLAG_TX_DLL = 0x10
    #: reserved
    PENDING_FLAG_TX_SYS = 0x20
    #: number of pending CAN messages in transmit buffer of firmware
    PENDING_FLAG_TX_FW = 0x40
    #: number of pending CAN messages in all receive buffers
    PENDING_FLAG_RX_ALL = PENDING_FLAG_RX_DLL | PENDING_FLAG_RX_SYS | PENDING_FLAG_RX_FW
    #: number of pending CAN messages in all transmit buffers
    PENDING_FLAG_TX_ALL = PENDING_FLAG_TX_DLL | PENDING_FLAG_TX_SYS | PENDING_FLAG_TX_FW
    #: number of pending CAN messages in all buffers
    PENDING_FLAG_ALL = PENDING_FLAG_RX_ALL | PENDING_FLAG_TX_ALL


class Mode(BYTE):
    """
    Specifies values for operation mode of a CAN channel.
    These values can be combined by OR operation with the method :meth:`UcanServer.init_can`.
    """

    #: normal operation mode (transmitting and receiving)
    MODE_NORMAL = 0
    #: listen only mode (receiving only, no ACK at CAN bus)
    MODE_LISTEN_ONLY = 1
    #: CAN messages which was sent will be received back with method :meth:`UcanServer.read_can_msg`
    MODE_TX_ECHO = 2
    #: reserved (not implemented in this version)
    MODE_RX_ORDER_CH = 4
    #: high resolution time stamps in received CAN messages (only available with STM derivatives)
    MODE_HIGH_RES_TIMER = 8


class VersionType(BYTE):
    """
    Specifies values for receiving the version information of several driver files.

    .. note:: This structure is only used internally.
    """

    #: version of the USB-CAN-library
    VER_TYPE_USER_LIB = 1
    #: equivalent to :attr:`VER_TYPE_USER_LIB`
    VER_TYPE_USER_DLL = 1
    #: version of USBCAN.SYS (not supported in this version)
    VER_TYPE_SYS_DRV = 2
    #: version of firmware in hardware (not supported, use method :meth:`UcanServer.get_fw_version`)
    VER_TYPE_FIRMWARE = 3
    #: version of UCANNET.SYS
    VER_TYPE_NET_DRV = 4
    #: version of USBCANLD.SYS
    VER_TYPE_SYS_LD = 5
    #: version of USBCANL2.SYS
    VER_TYPE_SYS_L2 = 6
    #: version of USBCANL3.SYS
    VER_TYPE_SYS_L3 = 7
    #: version of USBCANL4.SYS
    VER_TYPE_SYS_L4 = 8
    #: version of USBCANL5.SYS
    VER_TYPE_SYS_L5 = 9
    #: version of USBCANCP.CPL
    VER_TYPE_CPL = 10
